# pylint: disable=too-many-lines
"""Run subcommand implementation."""

from __future__ import annotations

import curses
import json
import logging
import re
import shlex
import shutil
import time
import uuid

from math import floor
from operator import itemgetter
from pathlib import Path
from queue import Queue
from typing import TYPE_CHECKING
from typing import Any

from ansible_navigator.action_base import ActionBase
from ansible_navigator.action_defs import RunStdoutReturn
from ansible_navigator.configuration_subsystem import to_effective
from ansible_navigator.configuration_subsystem import to_sources
from ansible_navigator.content_defs import ContentView
from ansible_navigator.content_defs import SerializationFormat
from ansible_navigator.runner import CommandAsync
from ansible_navigator.steps import Step
from ansible_navigator.ui_framework import CursesLine
from ansible_navigator.ui_framework import CursesLinePart
from ansible_navigator.ui_framework import CursesLines
from ansible_navigator.ui_framework import Interaction
from ansible_navigator.ui_framework import dict_to_form
from ansible_navigator.ui_framework import form_to_dict
from ansible_navigator.ui_framework import nonblocking_notification
from ansible_navigator.ui_framework import warning_notification
from ansible_navigator.utils.functions import check_playbook_type
from ansible_navigator.utils.functions import expand_path
from ansible_navigator.utils.functions import human_time
from ansible_navigator.utils.functions import is_jinja
from ansible_navigator.utils.functions import now_iso
from ansible_navigator.utils.functions import remove_ansi
from ansible_navigator.utils.functions import round_half_up
from ansible_navigator.utils.serialize import serialize_write_file

from . import _actions as actions
from . import run_action
from .stdout import Action as stdout_action


if TYPE_CHECKING:
    from collections.abc import Callable

    from ansible_navigator.app_public import AppPublic
    from ansible_navigator.configuration_subsystem.definitions import ApplicationConfiguration

COLUMN_IN_PROGRESS = "__in progress"

RESULT_TO_COLOR = [
    ("(?i)^failed$", 9),
    ("(?i)^ok$", 10),
    ("(?i)^ignored$", 13),
    ("(?i)^skipped$", 14),
    ("(?i)^in progress$", 8),
]


def get_color(word: str) -> int:
    """Retrieve color value matching the keyword.

    Args:
        word: Keyword to match color.

    Returns:
        Color value
    """
    return next(
        (x[1] for x in RESULT_TO_COLOR if re.match(x[0], word)),
        0,
    )


def _color_for_play(colname: str, colval: Any, entry: dict[str, Any]) -> tuple[int, int]:
    """Determine color and decoration for a play menu entry.

    Args:
        colname: The column name
        colval: The column value
        entry: The menu entry

    Returns:
        The color and decoration
    """
    color = 0
    decoration = 0
    if not colval:
        color = 8
    elif colname in ["__task_count", "__play_name", "__progress"]:
        failures = entry["__failed"] + entry["__unreachable"]
        if failures:
            color = 9
        elif entry["__ok"]:
            color = 10
        else:
            color = 8
    elif colname == "__changed":
        color = 11
    else:
        color = get_color(colname[2:])

    if colname == "__progress" and entry["__progress"].strip().lower() == "complete":
        decoration = curses.A_BOLD

    return color, decoration


def _color_for_task(colname: str, colval: Any, entry: dict[str, Any]) -> tuple[int, int]:
    """Determine color and decoration for a task menu entry.

    Args:
        colname: The column name
        colval: The column value
        entry: The menu entry

    Returns:
        The color and decoration
    """
    color = 0
    decoration = 0
    if entry["__result"].lower() == "in progress" or (
        colname in ["__result", "__host", "__number", "__task", "__task_action"]
    ):
        color = get_color(entry["__result"])
    elif colname == "__changed":
        color = 11 if colval is True else get_color(entry["__result"])
    elif colname == "__duration":
        color = 12
    return color, decoration


def color_menu(_colno: int, colname: str, entry: dict[str, Any]) -> tuple[int, int]:
    """Find matching color for word.

    Args:
        colname: The column name
        entry: The menu entry

    Returns:
        The color and decoration
    """
    colval = entry[colname]
    if "__play_name" in entry:
        return _color_for_play(colname, colval, entry)
    if "task" in entry:
        return _color_for_task(colname, colval, entry)
    return 0, 0


def content_heading(obj: Any, screen_w: int) -> CursesLines | None:
    """Create a heading for some piece of content showing.

    Args:
        obj: The content going to be shown
        screen_w: The current screen width

    Returns:
        The heading
    """
    if isinstance(obj, dict) and "task" in obj:
        detail = f"Play name: {obj['play']}:{obj['__number']}"

        line_1 = CursesLine(
            (CursesLinePart(column=0, string=detail, color=0, decoration=0),),
        )

        detail = f"Task name: {obj['task']}"
        line_2 = CursesLine(
            (CursesLinePart(column=0, string=detail, color=0, decoration=0),),
        )

        if obj["__changed"] is True:
            color = 11
            res = "CHANGED"
        else:
            color = next((x[1] for x in RESULT_TO_COLOR if re.match(x[0], obj["__result"])), 0)
            res = obj["__result"]

        if "res" in obj and "msg" in obj["res"]:
            msg = str(obj["res"]["msg"]).replace("\n", " ").replace("\r", "")
        else:
            msg = ""

        string = f"{res}: {obj['__host']} {msg}"
        string = string + (" " * (screen_w - len(string) + 1))
        line_3 = CursesLine(
            (CursesLinePart(column=0, string=string, color=color, decoration=curses.A_UNDERLINE),),
        )

        return CursesLines((line_1, line_2, line_3))
    return None


def filter_content_keys(obj: dict[Any, Any]) -> dict[Any, Any]:
    """Filter out some keys when showing collection content.

    Args:
        obj: The object from which keys should be removed

    Returns:
        The object with keys removed
    """
    return {k: v for k, v in obj.items() if not (k.startswith("_") or k.endswith("uuid"))}


PLAY_COLUMNS = [
    "__play_name",
    "__ok",
    "__changed",
    "__unreachable",
    "__failed",
    "__skipped",
    "__ignored",
    COLUMN_IN_PROGRESS,
    "__task_count",
    "__progress",
]

TASK_LIST_COLUMNS = [
    "__result",
    "__host",
    "__number",
    "__changed",
    "__task",
    "__task_action",
    "__duration",
]


@actions.register
class Action(ActionBase):
    # pylint: disable=too-many-instance-attributes
    """Run command implementation."""

    KEGEX = r"""(?x)
            ^
            (?P<run>r(?:un)?
            (\s(?P<params_run>.*))?)
            $"""

    def __init__(self, args: ApplicationConfiguration) -> None:
        """Initialize the ``:run`` action.

        Args:
            args: The current settings for the application
        """
        super().__init__(args=args, logger_name=__name__, name="run")

        self._subaction_type: str
        self._msg_from_plays: tuple[str | None, int | None] = (None, None)
        self._queue: Queue[dict[str, str]] = Queue()
        self.runner: CommandAsync
        self._runner_finished: bool
        self._auto_scroll = False
        #: Flag when the first message is received from runner
        self._first_message_received: bool = False

        self._plays = Step(
            name="plays",
            step_type="menu",
            columns=PLAY_COLUMNS,
            value=[],
            show_func=self._play_stats,
            select_func=self._task_list_for_play,
        )
        self._task_list_columns: list[str] = TASK_LIST_COLUMNS
        self._content_key_filter: Callable[[dict[Any, Any]], dict[Any, Any]] = filter_content_keys
        self._playbook_type: str = check_playbook_type(self._args.playbook)
        self._task_cache: dict[str, str] = {}
        """Task name storage from playbook_on_start using the task uuid as the key"""

    @property
    def mode(self) -> str:
        """Determine the mode and if playbook artifact creation is enabled.

        If so, run in interactive mode, but print stdout.

        Returns:
            If stdout and artifact creation, return a str statement as
            such, else return mode
        """
        if all(
            (
                self._args.mode == "stdout",
                self._args.playbook_artifact_enable,
                self._args.app != "replay",
            ),
        ):
            return "stdout_w_artifact"
        return self._args.mode

    def run_stdout(self) -> RunStdoutReturn:
        """Execute the ``run`` request for mode stdout.

        Returns:
            The return code from the runner invocation, along with a
            message to review the logs if not 0
        """
        if self._args.app == "replay":
            successful: bool = self._init_replay()
            if successful:
                return RunStdoutReturn(message="", return_code=0)
            return RunStdoutReturn(message="Please review the log for errors.", return_code=1)

        self._logger.debug("playbook requested in interactive mode")
        self._subaction_type = "playbook"
        self._logger = logging.getLogger(f"{__name__}_{self._subaction_type}")
        self._run_runner()
        while True:
            self._dequeue()
            if self.runner.finished:
                self._dequeue()
                if self._args.playbook_artifact_enable:
                    self.write_artifact()
                self._logger.debug("runner finished")
                break
            # Sleep briefly to prevent 100% CPU utilization
            # in mode stdout, the delay introduced by the curses key read is not present
            time.sleep(0.01)
        return_code = self.runner.ansible_runner_instance.rc
        if return_code != 0:
            return RunStdoutReturn(
                message="Please review the log for errors.",
                return_code=return_code,
            )
        return RunStdoutReturn(message="", return_code=return_code)

    def _initialize_subaction(self, interaction: Interaction) -> bool:
        """Initialize the run or replay subaction based on user interaction.

        Args:
            interaction: The interaction from the user

        Returns:
            True if initialization succeeded
        """
        if interaction.action.match.groupdict().get("run"):
            self._logger.debug("run requested in interactive mode")
            self._subaction_type = "run"
            str_uuid = str(uuid.uuid4())
            self._logger = logging.getLogger(f"{__name__}_{str_uuid[-4:]}")
            self._name = f"run_{str_uuid[-4:]}"
            return self._init_run()

        if interaction.action.match.groupdict().get("replay"):
            self._logger.debug("replay requested in interactive mode")
            self._subaction_type = "replay"
            self._name = "replay"
            self._logger = logging.getLogger(f"{__name__}_{self._subaction_type}")
            return self._init_replay()

        return False

    def run(self, interaction: Interaction, app: AppPublic) -> Interaction | None:
        """Run :run or :replay.

        Args:
            interaction: The interaction from the user
            app: The app instance

        Returns:
            The pending interaction or none
        """
        self._prepare_to_run(app, interaction)

        if not self._initialize_subaction(interaction):
            self._prepare_to_exit(interaction)
            return None

        self.steps.append(self._plays)

        # Show a notification until the first the first message from the queue is processed
        if self._subaction_type == "run":
            messages = ["Preparing for automation, please wait..."]
            notification = nonblocking_notification(messages=messages)
            interaction.ui.show_form(notification)
            while not self._first_message_received:
                self.update()

        while True:
            self.update()
            self._take_step()

            if not self.steps:
                if not self._runner_finished:
                    self._logger.error("Can not step back while playbook in progress, :q! to exit")
                    self.steps.append(self._plays)
                    continue
                self._logger.debug(
                    "No steps remaining for '%s' returning to calling app",
                    self._name,
                )
                break

            if self.steps.current.name == "quit":
                result = self._handle_quit(interaction)
                if result is not None:
                    return result

        self._prepare_to_exit(interaction)
        return None

    def _handle_quit(self, interaction: Interaction) -> Interaction | None:
        """Handle quit step logic during the run loop.

        Args:
            interaction: The interaction from the user

        Returns:
            The quit interaction if exiting, None to continue the loop
        """
        if self._args.app == "replay":
            self._prepare_to_exit(interaction)
            return self.steps.current
        done = self._prepare_to_quit(self.steps.current)
        if done:
            self._prepare_to_exit(interaction)
            return self.steps.current
        self.steps.back_one()
        return None

    def _validate_and_prompt_playbook(self) -> bool:
        """Validate the playbook path and prompt the user if invalid.

        Returns:
            True if the playbook is valid or the user provided a valid one
        """
        if isinstance(self._args.playbook, str):
            playbook_valid = Path(self._args.playbook).exists()
        else:
            playbook_valid = False

        if playbook_valid:
            return True

        populated_form = self._prompt_for_playbook()
        if populated_form["cancelled"]:
            return False

        new_cmd = ["run"]
        new_cmd.append(populated_form["fields"]["playbook"]["value"])

        if populated_form["fields"]["cmdline"]["value"]:
            new_cmd.extend(shlex.split(populated_form["fields"]["cmdline"]["value"]))

        # Parse as if provided from the cmdline
        return self._update_args(new_cmd)

    def _init_run(self) -> bool:
        """In the case of :run, check the user input.

        Returns:
            True
        """
        # Ensure the playbook and inventory are valid

        args_updated = self._update_args(
            ["run"] + shlex.split(self._interaction.action.match.groupdict()["params_run"] or ""),
        )
        if not args_updated:
            return False

        if self._playbook_type == "file" and not self._validate_and_prompt_playbook():
            return False

        self._run_runner()
        self._logger.info("Run initialized and playbook started.")
        return True

    def _validate_artifact_file(self) -> str | None:
        """Validate the artifact file path and prompt if needed.

        Returns:
            The resolved artifact file path, or None if validation failed
        """
        artifact_file = self._args.playbook_artifact_replay

        if isinstance(self._args.playbook_artifact_replay, str):
            artifact_valid = Path(self._args.playbook_artifact_replay).exists()
        else:
            artifact_valid = False

        if not artifact_valid:
            if self.mode == "interactive":
                populated_form = self._prompt_for_artifact(artifact_file=artifact_file)
                if populated_form["cancelled"]:
                    return None
                artifact_file = populated_form["fields"]["artifact_file"]["value"]
            else:
                self._logger.error("Artifact file '%s' not found", artifact_file)
                return None

        return artifact_file

    def _load_artifact_data(self, data: dict[str, Any]) -> bool:
        """Load artifact data based on version and current mode.

        Args:
            data: The parsed artifact data

        Returns:
            True if the data was loaded successfully
        """
        version = data.get("version", "")
        if not version.startswith(("1.", "2.")):
            self._logger.error(
                "Incompatible artifact version, got '%s', compatible = '1.y.z'",
                version,
            )
            return False

        try:
            stdout = data["stdout"]
            if self.mode == "interactive":
                self._plays.value = data["plays"]
                self._interaction.ui.update_status(data["status"], data["status_color"])
                self.stdout = stdout
            else:
                for line in data["stdout"]:
                    if self._args.display_color is True:
                        print(line)
                    else:
                        print(remove_ansi(line))
        except KeyError as exc:
            self._logger.debug("missing keys from artifact file")
            self._logger.debug("error was: %s", str(exc))
            return False

        return True

    def _init_replay(self) -> bool:
        """In the case of :replay, replay the artifact.

        Check for a version and process artifact file.

        Returns:
            True if replay completes, False if there is an error
        """
        self._logger.debug("Starting replay artifact request with mode %s", self.mode)

        if self.mode == "interactive":
            args_updated = self._update_args(
                ["replay"]
                + shlex.split(self._interaction.action.match.groupdict()["params_replay"] or ""),
            )
            if not args_updated:
                return False

        artifact_file = self._validate_artifact_file()
        if artifact_file is None:
            return False

        try:
            with Path(artifact_file).open(encoding="utf-8") as fh:
                data = json.load(fh)
        except json.JSONDecodeError as exc:
            self._logger.debug("json decode error: %s", str(exc))
            self._logger.exception("Unable to parse artifact file")
            return False

        if not self._load_artifact_data(data):
            return False

        self._runner_finished = True
        self._logger.debug("Completed replay artifact request with mode %s", self.mode)
        return True

    def _prompt_for_artifact(self, artifact_file: str) -> dict[Any, Any]:
        """Prompt for a valid artifact file.

        Args:
            artifact_file: Artifact file

        Returns:
            Dict with artifact detail entries
        """
        if not isinstance(artifact_file, str):
            artifact_file = ""

        form_dict: dict[str, Any] = {
            "title": "Artifact file not found, please confirm the following",
            "fields": [],
        }
        form_field = {
            "name": "artifact_file",
            "prompt": "Path to artifact file",
            "type": "text_input",
            "validator": {"name": "valid_file_path"},
            "pre_populate": artifact_file,
        }
        form_dict["fields"].append(form_field)
        form = dict_to_form(form_dict)
        self._interaction.ui.show_form(form)
        populated_form = form_to_dict(form, key_on_name=True)
        return populated_form

    def _prompt_for_playbook(self) -> dict[Any, Any]:
        """Pre-populate a form to confirm the playbook details.

        Returns:
            Dict with playbook and inventory detail entries
        """
        self._logger.debug("Inventory/Playbook not set, provided, or valid, prompting")

        playbook = self._args.playbook if isinstance(self._args.playbook, str) else ""

        cmdline = " ".join(self._args.cmdline) if isinstance(self._args.cmdline, list) else ""

        form_dict: dict[str, Any] = {
            "title": "Playbook not found, please confirm the following",
            "fields": [],
        }
        form_field = {
            "name": "playbook",
            "pre_populate": playbook,
            "prompt": "Path to playbook",
            "type": "text_input",
            "validator": {"name": "valid_file_path"},
        }
        form_dict["fields"].append(form_field)

        form_field = {
            "name": "cmdline",
            "pre_populate": cmdline,
            "prompt": "Additional command line parameters",
            "type": "text_input",
            "validator": {"name": "none"},
        }
        form_dict["fields"].append(form_field)
        form = dict_to_form(form_dict)
        self._interaction.ui.show_form(form)
        populated_form = form_to_dict(form, key_on_name=True)
        return populated_form

    def _handle_step_display(self) -> Interaction | None:
        """Handle display logic for a Step on the stack.

        Returns:
            The interaction result from the UI, or None
        """
        if self.steps.current.show_func:
            self.steps.current.show_func()

        if self.steps.current.type == "menu":
            return self._handle_menu_step()

        if self.steps.current.type == "content":
            return self._interaction.ui.show(
                obj=self.steps.current.value,
                index=self.steps.current.index,
                content_heading=content_heading,
                filter_content_keys=self._content_key_filter,
            )
        return None

    def _handle_menu_step(self) -> Interaction | None:
        """Handle display and auto-scroll logic for a menu step.

        Returns:
            The interaction result from the UI, or None
        """
        new_scroll = len(self.steps.current.value)
        if self._auto_scroll:
            self._interaction.ui.scroll(new_scroll)

        result = self._interaction.ui.show(
            obj=self.steps.current.value,
            columns=self.steps.current.columns,
            color_menu_item=color_menu,
        )

        if self._interaction.ui.scroll() < new_scroll and self._auto_scroll:
            self._logger.debug("auto_scroll disabled")
            self._auto_scroll = False
        elif self._interaction.ui.scroll() >= new_scroll and not self._auto_scroll:
            self._logger.debug("auto_scroll enabled")
            self._auto_scroll = True

        return result

    def _take_step(self) -> None:
        """Run the current step on the stack."""
        result = None
        if isinstance(self.steps.current, Interaction):
            result = run_action(self.steps.current.name, self.app, self.steps.current)
        elif isinstance(self.steps.current, Step):
            result = self._handle_step_display()

        if result is None:
            self.steps.back_one()
        else:
            self.steps.append(result)

    def _run_runner(self) -> None:
        """Spin up runner.

        Raises:
            RuntimeError: If no ansible-playbook executable
        """
        executable_cmd: str | None

        mode = "interactive" if self.mode == "stdout_w_artifact" else self.mode

        if isinstance(self._args.set_environment_variable, dict):
            set_env_vars = {**self._args.set_environment_variable}
        else:
            set_env_vars = {}

        if self._args.display_color is False:
            set_env_vars["ANSIBLE_NOCOLOR"] = "1"

        kwargs = {
            "container_engine": self._args.container_engine,
            "host_cwd": str(Path.cwd()),
            "execution_environment_image": self._args.execution_environment_image,
            "execution_environment": self._args.execution_environment,
            "inventory": self._args.inventory,
            "navigator_mode": mode,
            "pass_environment_variable": self._args.pass_environment_variable,
            "set_environment_variable": set_env_vars,
            "private_data_dir": self._args.ansible_runner_artifact_dir,
            "rotate_artifacts": self._args.ansible_runner_rotate_artifacts_count,
            "timeout": self._args.ansible_runner_timeout,
        }

        if isinstance(self._args.playbook, str):
            kwargs.update({"playbook": self._args.playbook})

        if isinstance(self._args.execution_environment_volume_mounts, list):
            kwargs.update(
                {"container_volume_mounts": self._args.execution_environment_volume_mounts},
            )

        if isinstance(self._args.container_options, list):
            kwargs.update({"container_options": self._args.container_options})

        if self._args.execution_environment:
            executable_cmd = "ansible-playbook"
        else:
            executable_cmd = shutil.which("ansible-playbook")
            if not executable_cmd:
                msg = "'ansible-playbook' executable not found"
                self._logger.error(msg)
                raise RuntimeError(msg)

        pass_through_arg = []
        if self._args.help_playbook is True:
            pass_through_arg.append("--help")
        if isinstance(self._args.cmdline, list):
            pass_through_arg.extend(self._args.cmdline)
        kwargs.update({"cmdline": pass_through_arg})

        self.runner = CommandAsync(
            executable_cmd=executable_cmd,
            queue=self._queue,
            write_job_events=self._args.ansible_runner_write_job_events,
            **kwargs,
        )
        self.runner.run()
        self._runner_finished = False
        self._logger.debug("runner requested to start")

    def _dequeue(self) -> None:
        """Drain the runner queue."""
        drain_count = 0
        while not self._queue.empty():
            if not self._first_message_received:
                self._first_message_received = True
            message = self._queue.get()
            self._handle_message(message)
            drain_count += 1
        if drain_count:
            self._logger.debug("Drained %s events", drain_count)

    def _handle_runner_on_start(
        self,
        play: dict[str, Any],
        event_data: dict[str, Any],
    ) -> None:
        """Handle a runner_on_start event by adding a new in-progress task.

        Args:
            play: The parent play data
            event_data: The event data from the runner
        """
        try:
            previous_name = self._task_cache[event_data["task_uuid"]]
            use_previous = not is_jinja(previous_name)
        except KeyError:
            use_previous = False

        event_data.update(
            {
                "__changed": "unknown",
                "__duration": "Pending",
                "__host": event_data["host"],
                "__number": len(play["tasks"]),
                "__result": "In progress",
                "__task_action": event_data["task_action"],
                "__task": previous_name if use_previous else event_data["task"],
            },
        )
        play["tasks"].append(event_data)

    def _handle_runner_on_finish(
        self,
        play: dict[str, Any],
        event_data: dict[str, Any],
        runner_event: str,
    ) -> None:
        """Handle a runner task-finished event by updating the task with results.

        Args:
            play: The parent play data
            event_data: The event data from the runner
            runner_event: The runner event type (ok, skipped, failed, etc.)
        """
        match = ("task_uuid", "host")
        try:
            task = next(
                t for t in play["tasks"] if itemgetter(*match)(t) == itemgetter(*match)(event_data)
            )
        except StopIteration:
            self._logger.warning("Task event without parent task")
            return

        # Get the duration
        if isinstance(event_data["duration"], int | float):
            duration = human_time(seconds=round_half_up(event_data["duration"]))
        else:
            self._logger.debug(
                "Task duration for '%s' was type '%s', set to 0",
                event_data["task"],
                type(event_data["duration"]),
            )
            duration = human_time(seconds=0)

        # Replace failed with ignored
        modify_result = runner_event == "failed" and event_data["ignore_errors"]

        event_data.update(
            {
                "__changed": event_data.get("res", {}).get("changed", False),
                "__duration": duration,
                "__result": ("ignored" if modify_result else runner_event).capitalize(),
            },
        )

        # Find the best name for the task
        name_changed = task["__task"] != event_data["task"]
        no_longer_templated = is_jinja(task["__task"]) and not is_jinja(event_data["task"])
        changed_and_not_templated = name_changed and not is_jinja(event_data["task"])
        if no_longer_templated or changed_and_not_templated:
            event_data["__task"] = event_data["task"]

        task.update(event_data)

    def _handle_message(self, message: dict[str, Any]) -> None:
        """Handle a runner message.

        Args:
            message (dict): The message from runner
        """
        # Collect any stdout
        if message.get("stdout"):
            self.stdout.extend(message["stdout"].splitlines())
            if self.mode == "stdout_w_artifact":
                print(message["stdout"])

        # Get the event data
        try:
            event = message["event"]
            event_data = message["event_data"]
        except KeyError:
            error = f"Unhandled message from runner queue, discarded: {message}"
            self._logger.critical(error)
            return

        # Handle verbose and error messages
        if event in ["verbose", "error"]:
            if "ERROR!" in message["stdout"]:
                self._msg_from_plays = ("ERROR", 9)
                if self.mode == "interactive":
                    self._notify_error(message["stdout"])
            elif "WARNING" in message["stdout"]:
                self._msg_from_plays = ("WARNINGS", 13)
            return

        if event == "playbook_on_play_start":
            event_data["__play_name"] = event_data["name"]
            event_data["tasks"] = []
            self._plays.value.append(event_data)
            return

        if event == "playbook_on_task_start":
            self._task_cache[event_data["task_uuid"]] = event_data["task"]
            return

        # Only runner on_* events are relevant now
        try:
            prefix, runner_event = event.rsplit("_", 1)
            assert prefix == "runner_on"  # noqa:S101
            assert runner_event in ("ok", "skipped", "start", "unreachable", "failed")  # noqa:S101
        except (AssertionError, ValueError):
            return

        # Find the parent play of the task
        try:
            play = next(p for p in self._plays.value if p["uuid"] == event_data["play_uuid"])
        except StopIteration:
            self._logger.warning("Playbook event without parent play")
            return

        if runner_event == "start":
            self._handle_runner_on_start(play, event_data)
        else:
            self._handle_runner_on_finish(play, event_data, runner_event)

    def _play_stats(self) -> None:
        """Calculate the play's stats based on it's tasks."""
        for idx, play in enumerate(self._plays.value):
            total = [
                "__ok",
                "__skipped",
                "__failed",
                "__unreachable",
                "__ignored",
                COLUMN_IN_PROGRESS,
            ]
            self._plays.value[idx].update(
                {
                    tot: len([t for t in play["tasks"] if t["__result"].lower() == tot[2:]])
                    for tot in total
                },
            )
            self._plays.value[idx]["__changed"] = len(
                [t for t in play["tasks"] if t["__changed"] is True],
            )
            task_count = len(play["tasks"])
            self._plays.value[idx]["__task_count"] = task_count
            completed = task_count - self._plays.value[idx][COLUMN_IN_PROGRESS]
            if completed:
                new = floor(completed / task_count * 100)
                current = self._plays.value[idx].get("__percent_complete", 0)
                self._plays.value[idx]["__percent_complete"] = max(new, current)
                self._plays.value[idx]["__progress"] = str(max(new, current)) + "%"
            else:
                self._plays.value[idx]["__progress"] = "0%"

    def _prepare_to_quit(self, interaction: Interaction) -> bool:
        """Pre-quit tasks.

        Args:
            interaction: The quit interaction

        Returns:
            A bool indicating whether or not it's safe to exit
        """
        self.update()
        if self.runner is not None and not self.runner.finished:
            if interaction.action.match.groupdict()["exclamation"]:
                self._logger.debug("shutting down runner")
                self.runner.cancelled = True
                while not self.runner.finished:
                    pass  # busy-wait until the runner finishes
                self.write_artifact()
                return True
            self._logger.warning("Quit requested but playbook running, try q! or quit!")
            return False
        self._logger.debug("runner not running")
        return True

    def _task_list_for_play(self) -> Step:
        """Generate a menu of tasks for the currently selected play.

        Returns:
            The menu step
        """
        value = self.steps.current.selected["tasks"]
        step = Step(
            name="task_list",
            step_type="menu",
            columns=self._task_list_columns,
            select_func=self._task_from_task_list,
            value=value,
        )
        return step

    def _task_from_task_list(self) -> Step:
        """Generate task content for the selected task.

        Returns:
            Content which shows a task
        """
        value = self.steps.current.value
        index = self.steps.current.index
        step = Step(name="task", step_type="content", index=index, value=value)
        return step

    def update(self) -> None:
        """Drain the queue, set the status and write the artifact if needed."""
        # let the calling app update as well
        self._calling_app.update()

        if hasattr(self, "runner"):
            self._dequeue()
            self._set_status()

            if self.runner.finished and not self._runner_finished:
                self._logger.debug("runner finished")
                self._logger.info("Playbook complete")
                self.write_artifact()
                self._runner_finished = True
                self._notify_no_tasks_redirect()

    def _get_status(self) -> tuple[str, int]:
        """Get the runner status and color for status message.

        Returns:
            status string, status color
        """
        status = ""
        status_color = 0
        if self.runner.status:
            if self.runner and self.runner.finished and self.runner.status:
                status = self.runner.status
                if self.runner.status == "failed":
                    status_color = 9
                else:
                    status_color = self._msg_from_plays[1] or 10
            elif self._msg_from_plays[0] is not None and self._msg_from_plays[1] is not None:
                status = self._msg_from_plays[0]
                status_color = self._msg_from_plays[1]
            else:
                status = self.runner.status
                status_color = 10
        return status, status_color

    def _set_status(self) -> None:
        """Set the UI status."""
        status, status_color = self._get_status()
        self._interaction.ui.update_status(status, status_color)

    def write_artifact(self, filename: str | None = None) -> None:
        """Write the artifact.

        Args:
            filename (str): The file to write to
        """
        if filename or self._args.playbook_artifact_enable is True:
            status, status_color = self._get_status()
            playbook = self._args.playbook
            if self._playbook_type == "fqcn" and len(self._plays.value) > 0:
                playbook = next(k["playbook"] for k in self._plays.value)
            filename = filename or self._args.playbook_artifact_save_as
            filename = filename.format(
                playbook_dir=Path(playbook).parent,
                playbook_name=Path(playbook).stem,
                playbook_status=status,
                time_stamp=now_iso(self._args.time_zone),
            )
            self._logger.debug("Formatted artifact file name set to %s", filename)
            filename = str(expand_path(filename))
            self._logger.debug("Resolved artifact file name set to %s", filename)

            try:
                Path(Path(filename).parent).mkdir(parents=True, exist_ok=True)
                artifact = {
                    "version": "2.0.0",
                    "plays": self._plays.value,
                    "stdout": self.stdout,
                    "status": status,
                    "status_color": status_color,
                    "settings_entries": to_effective(self._args),
                    "settings_sources": to_sources(self._args),
                }
                serialize_write_file(
                    content=artifact,
                    content_view=ContentView.NORMAL,
                    file_mode="w",
                    file=Path(filename),
                    serialization_format=SerializationFormat.JSON,
                )
                self._logger.info("Saved artifact as %s", filename)

            except OSError as exc:
                error = (
                    f"Saving the artifact file failed, resulted in the following error: f{exc!s}"
                )
                self._logger.exception(error)

    def rerun(self) -> None:
        """Rerun the current playbook.

        Since we're not reinstating run, drain the queue,
        clear the steps, reset the index, etc.
        """
        if self._subaction_type == "run":
            if self.runner.finished:
                self._plays.value = []
                self._plays.index = None
                self._msg_from_plays = (None, None)
                self._queue.queue.clear()
                self.stdout = []
                self._run_runner()
                self.steps.clear()
                self.steps.append(self._plays)
                self._logger.debug("Playbook rerun triggered")
            else:
                self._logger.warning("Playbook rerun ignored, current playbook not complete")
        elif self._subaction_type == "replay":
            self._logger.error("No rerun available when artifact is loaded")
        else:
            self._logger.error("sub-action type '%s' is invalid", self._subaction_type)

    def _notify_error(self, message: str) -> None:
        """Show a blocking warning.

        Args:
            message: Message for warning
        """
        warn_msg = ["Errors were encountered while running the playbook:"]
        messages = remove_ansi(message).splitlines()
        messages[-1] += "..."
        warn_msg.extend(messages)
        warn_msg += ["[HINT] To see the full error message try ':stdout'"]
        warn_msg += ["[HINT] After it's fixed, try to ':rerun' the playbook"]
        warning = warning_notification(warn_msg)
        self._interaction.ui.show_form(warning)

    def _notify_no_tasks_redirect(self) -> None:
        """In the case the playbook finished but without tasks, show a warning, send to stdout."""
        total_tasks = sum(len(play["tasks"]) for play in self._plays.value)
        # At least one task, no need to redirect
        if total_tasks:
            return

        message = ["The playbook completed without tasks. Redirecting to ':stdout' for review."]
        warning = warning_notification(message)

        self._interaction.ui.show_form(warning)
        self.steps.append(Interaction(name="stdout", action=stdout_action, ui=self._interaction.ui))  # type: ignore



"""Collections subcommand implementation."""

from __future__ import annotations

import curses
import json
import shlex
import sys

from copy import deepcopy
from json.decoder import JSONDecodeError
from pathlib import Path
from typing import TYPE_CHECKING
from typing import Any

from ansible_navigator.action_base import ActionBase
from ansible_navigator.action_defs import RunStdoutReturn
from ansible_navigator.content_defs import ContentFormat
from ansible_navigator.runner import Command
from ansible_navigator.steps import Step
from ansible_navigator.ui_framework import CursesLine
from ansible_navigator.ui_framework import CursesLinePart
from ansible_navigator.ui_framework import CursesLines
from ansible_navigator.ui_framework import Interaction
from ansible_navigator.ui_framework import nonblocking_notification
from ansible_navigator.ui_framework import warning_notification
from ansible_navigator.utils.functions import path_is_relative_to
from ansible_navigator.utils.functions import remove_dbl_un
from ansible_navigator.utils.key_value_store import KeyValueStore
from ansible_navigator.utils.print import print_to_stdout

from . import _actions as actions
from . import run_action


if TYPE_CHECKING:
    from ansible_navigator.app_public import AppPublic
    from ansible_navigator.configuration_subsystem.definitions import ApplicationConfiguration


def color_menu(colno: int, colname: str, entry: dict[str, Any]) -> tuple[int, int]:
    """Provide a color for a collections menu entry in one column.

    Args:
        colno: The column number
        colname: The column name
        entry: The menu entry

    Returns:
        The color and decoration
    """
    if entry.get("__shadowed") is True:
        return 8, 0
    if entry.get("__deprecated") == "True":
        return 9, 0
    return 2, 0


def content_heading(obj: Any, screen_w: int) -> CursesLines | None:
    """Create a heading for collection content.

    Args:
        obj: The content going to be shown
        screen_w: The current screen width

    Returns:
        The heading
    """
    name = f"Image: {obj['full_name']}"
    description = f"Description: {obj['__description']}"
    padding = " " * (screen_w - len(description) + 1)
    line_1_part_1 = CursesLinePart(
        column=0,
        string=name,
        color=0,
        decoration=curses.A_NORMAL,
    )
    line_2_part_1 = CursesLinePart(
        column=0,
        string=description + padding,
        color=0,
        decoration=curses.A_UNDERLINE,
    )
    return CursesLines((CursesLine((line_1_part_1,)), CursesLine((line_2_part_1,))))


def filter_content_keys(obj: dict[Any, Any]) -> dict[Any, Any]:
    """Filter out some keys when showing collection content.

    Args:
        obj: The object from which keys should be removed

    Returns:
        The object with keys removed
    """
    return {k: v for k, v in obj.items() if not k.startswith("__")}


@actions.register
class Action(ActionBase):
    """Collections subcommand implementation."""

    KEGEX = r"^collections(\s(?P<params>.*))?$"

    def __init__(self, args: ApplicationConfiguration) -> None:
        """Initialize the ``collections`` action.

        Args:
            args: The current settings for the application
        """
        super().__init__(args=args, logger_name=__name__, name="collections")
        self._adjacent_collection_dir: str
        self._collection_cache: KeyValueStore
        self._collection_cache_path: str
        self._collection_scanned_paths: list[str] = []
        self._collections: list[Any] = []
        self._stats: dict[str, Any] = {}

    def update(self) -> None:
        """Request calling app update, no collection update is required."""
        self._calling_app.update()

    def run(self, interaction: Interaction, app: AppPublic) -> Interaction | None:
        """Execute the ``collections`` request for mode interactive.

        Args:
            interaction: The interaction from the user
            app: The app instance

        Returns:
            The pending
            :class:`~ansible_navigator.ui_framework.ui.Interaction` or
            :data:`None`
        """
        self._logger.debug("collections requested")
        self._prepare_to_run(app, interaction)
        self.stdout = self._calling_app.stdout

        notification = nonblocking_notification(
            messages=[
                "Collecting collection content, this may take a minute the first time...",
            ],
        )
        interaction.ui.show_form(notification)

        params = [self._name] + shlex.split(
            self._interaction.action.match.groupdict()["params"] or "",
        )

        args_updated = self._update_args(params=params, attach_cdc=True)
        if not args_updated:
            self._prepare_to_exit(interaction)
            return None

        if not isinstance(self._args.internals.collection_doc_cache, KeyValueStore):
            notification = warning_notification(
                messages=[
                    "Something has gone really wrong, the collection document cache is not",
                    "available.  This should not have happened. Please log an issue, and",
                    "include the contents of the log file.",
                ],
            )
            interaction.ui.show_form(notification)
            self._prepare_to_exit(interaction)
            return None

        self._collection_cache = self._args.internals.collection_doc_cache
        self._collection_cache_path = self._args.collection_doc_cache_path

        self._run_runner()

        if not self._collections:
            self._prepare_to_exit(interaction)
            self.notify_none()
            return None

        self.steps.append(self._build_main_menu())

        while True:
            self.update()
            self._take_step()

            if not self.steps:
                break

            if self.steps.current.name == "quit":
                return self.steps.current

        self._prepare_to_exit(interaction)
        return None

    def run_stdout(self) -> RunStdoutReturn:
        """Execute the ``collection`` request for mode stdout.

        Returns:
            The return code or 1. If the response from the runner
            invocation is None, indicates there is no console output to
            display, so assume an issue and return 1 along with a
            message to review the logs.
        """
        self._logger.debug("collection requested in stdout mode")

        args_updated = self._update_args(params=[], attach_cdc=True)
        if not args_updated or not isinstance(
            self._args.internals.collection_doc_cache,
            KeyValueStore,
        ):
            msg = (
                "Failed to create collections cache, "
                "Please review the ansible-navigator log file for errors."
            )
            return RunStdoutReturn(message=msg, return_code=1)

        self._collection_cache = self._args.internals.collection_doc_cache
        self._collection_cache_path = self._args.collection_doc_cache_path

        self._run_runner()
        if not self._collections:
            msg = (
                "Failed to catalog collections, "
                "Please review the ansible-navigator log file for errors."
            )
            return RunStdoutReturn(message=msg, return_code=1)

        collections_info = self._parse_collection_info_stdout()

        print_to_stdout(
            content=collections_info,
            content_format=getattr(ContentFormat, self._args.format.upper()),
            use_color=self._args.display_color,
        )
        return RunStdoutReturn(message="", return_code=0)

    def notify_failed(self) -> None:
        """Notify collection cataloging failed."""
        msgs = ["humph. Something went really wrong while cataloging collections."]
        msgs.append("Details have been added to the log file")
        closing = ["[HINT] Please log an issue about this one, it shouldn't have happened"]

        # Only show UI notification in interactive mode
        if self._args.mode == "interactive" and hasattr(self, "_interaction"):
            warning = warning_notification(messages=msgs + closing)
            self._interaction.ui.show_form(warning)
        else:
            # In stdout mode, just log the error
            for msg in msgs + closing:
                self._logger.error(msg)

    def notify_none(self) -> None:
        """Notify no collections were found."""
        msgs = ["humph. no collections were found in the following paths:"]
        paths = []
        for path in self._collection_scanned_paths:
            if path.startswith(str(self._args.internals.cache_path)):
                continue
            if self._args.execution_environment:
                if path.startswith(self._adjacent_collection_dir):
                    paths.append(f"- {path} (bind_mount)")
                else:
                    paths.append(f"- {path} (contained)")
            else:
                paths.append(f"- {path}")
        closing = ["[HINT] Try installing some or try a different execution environment"]
        warning = warning_notification(messages=msgs + paths + closing)
        self._interaction.ui.show_form(warning)

    def _take_step(self) -> None:
        """Take a step based on the current step or step back."""
        result = None
        if isinstance(self.steps.current, Interaction):
            result = run_action(self.steps.current.name, self.app, self.steps.current)
        elif isinstance(self.steps.current, Step):
            if self.steps.current.show_func:
                current_index = self.steps.current.index
                self.steps.current.show_func()
                self.steps.current.index = current_index

            if self.steps.current.type == "menu":
                result = self._interaction.ui.show(
                    obj=self.steps.current.value,
                    columns=self.steps.current.columns,
                    color_menu_item=color_menu,
                )
            elif self.steps.current.type == "content":
                result = self._interaction.ui.show(
                    obj=self.steps.current.value,
                    index=self.steps.current.index,
                    content_heading=content_heading,
                    filter_content_keys=filter_content_keys,
                )

        if result is None:
            self.steps.back_one()
        else:
            self.steps.append(result)

    def _build_main_menu(self) -> Step:
        """Build the menu of collections.

        Returns:
            The collections menu definition
        """
        if self._args.execution_environment:
            columns = ["__name", "__version", "__shadowed", "__type", "path"]
        else:
            columns = ["__name", "__version", "__shadowed", "path"]

        return Step(
            name="all_collections",
            columns=columns,
            select_func=self._build_collection_content_menu,
            step_type="menu",
            value=self._collections,
        )

    def _process_plugin_for_menu(
        self,
        plugin_checksum: str,
        details: dict[str, Any],
        selected_collection: dict[str, Any],
        collection_name: str,
    ) -> dict[str, Any] | None:
        """Process a single plugin entry for the collection content menu.

        Args:
            plugin_checksum: The checksum key for the plugin in the cache
            details: The plugin details including type information
            selected_collection: The parent collection data
            collection_name: The display column name for the collection

        Returns:
            The processed plugin dict, or None if it could not be loaded
        """
        try:
            plugin_json = self._collection_cache[plugin_checksum]
            loaded = json.loads(plugin_json)

            plugin = loaded["plugin"]
            if plugin["doc"] is None:
                return None

            if "name" in plugin["doc"]:
                short_name = plugin["doc"]["name"]
            else:
                short_name = plugin["doc"][details["type"]]
            plugin[collection_name] = short_name
            plugin["full_name"] = f"{selected_collection['known_as']}.{short_name}"
            plugin["__type"] = details["type"]
            plugin["collection_info"] = selected_collection["collection_info"]
            plugin["collection_info"]["name"] = selected_collection["known_as"]
            plugin["collection_info"]["shadowed_by"] = selected_collection["hidden_by"]
            plugin["collection_info"]["path"] = selected_collection["path"]

            plugin["__added"] = plugin["doc"].get("version_added")
            plugin["__description"] = plugin["doc"]["short_description"]

            runtime_section = "modules" if details["type"] == "module" else details["type"]
            plugin["__deprecated"] = "False"
            try:
                routing_info = selected_collection["runtime"]["plugin_routing"]
                runtime_info = routing_info[runtime_section][short_name]
                plugin["additional_information"] = runtime_info
                if "deprecation" in runtime_info:
                    plugin["__deprecated"] = "True"
            except KeyError:
                plugin["additional_information"] = {}
        except (KeyError, JSONDecodeError) as exc:
            self._logger.exception("error loading plugin doc %s", details)
            self._logger.debug("error was %s", str(exc))
            return None
        else:
            return plugin

    def _process_roles_for_menu(
        self,
        roles: list[dict[str, Any]],
        collection_name: str,
    ) -> list[dict[str, Any]]:
        """Process roles for the collection content menu.

        Args:
            roles: The list of role dicts from the collection
            collection_name: The display column name for the collection

        Returns:
            The list of processed role dicts ready for menu display
        """
        processed = []
        for role in roles:
            role[collection_name] = role["short_name"]
            try:
                role["__description"] = role["info"]["galaxy_info"]["description"]
            except KeyError:
                role["__description"] = ""
            role["__deprecated"] = "Unknown"
            role["__added"] = "Unknown"
            role["__type"] = "role"
            processed.append(role)
        return processed

    def _build_collection_content_menu(self) -> Step:
        """Build the menu of plugins.

        Returns:
            The plugin menu definition
        """
        self._collection_cache.open_()
        selected_collection = self._collections[self.steps.current.index]
        collection_name = f"__{selected_collection['known_as']}"
        collection_contents = []
        for plugin_checksum, details in selected_collection["plugin_checksums"].items():
            plugin = self._process_plugin_for_menu(
                plugin_checksum,
                details,
                selected_collection,
                collection_name,
            )
            if plugin is not None:
                collection_contents.append(plugin)

        self._collection_cache.close()

        collection_contents.extend(
            self._process_roles_for_menu(selected_collection["roles"], collection_name),
        )

        collection_contents = sorted(collection_contents, key=lambda i: i[collection_name])

        return Step(
            name="all_collection_content",
            columns=[collection_name, "__type", "__added", "__deprecated", "__description"],
            select_func=self._build_collection_content,
            step_type="menu",
            value=collection_contents,
        )

    def _build_collection_content(self) -> Step:
        """Build the content for one plugin.

        Returns:
            The plugin's content
        """
        return Step(
            name="collection_content",
            step_type="content",
            value=self.steps.current.value,
            index=self.steps.current.index,
        )

    def _setup_ee_pythonpath(
        self,
        set_environment_variable: dict[str, Any],
    ) -> None:
        """Configure PYTHONPATH for execution environment runs.

        Injects the navigator utils mount path into PYTHONPATH so that
        the key_value_store module is accessible inside the EE.

        Args:
            set_environment_variable: The environment variable dict to update in place
        """
        ee_navigator_utils_mount = "/opt/ansible_navigator_utils"
        if "PYTHONPATH" in set_environment_variable:
            set_environment_variable["PYTHONPATH"].append(
                f":{ee_navigator_utils_mount}",
            )
        else:
            set_environment_variable["PYTHONPATH"] = f"${{PYTHONPATH}}:{ee_navigator_utils_mount}"
        self._logger.debug(
            "Execution Environment's PYTHONPATH is set to: %s",
            set_environment_variable["PYTHONPATH"],
        )

    def _setup_ee_volume_mounts(
        self,
        kwargs: dict[str, Any],
        playbook_dir: str,
        cache_path: Path,
    ) -> str:
        """Build container volume mounts and return the python executable path for EE runs.

        Args:
            kwargs: The runner kwargs dict to update with volume mounts
            playbook_dir: The playbook directory path
            cache_path: The application cache path

        Returns:
            The python executable path to use inside the EE
        """
        self._logger.debug("running collections command with execution environment enabled")
        python_exec_path = f"{cache_path}/python_latest.sh"
        utils_lib = Path(__file__).parent / ".." / "utils"

        container_volume_mounts = [
            f"{cache_path}:{cache_path}",
            f"{utils_lib!s}:/opt/ansible_navigator_utils",
        ]
        if Path(self._adjacent_collection_dir).exists():
            container_volume_mounts.append(
                f"{self._adjacent_collection_dir}:{self._adjacent_collection_dir}:z",
            )

        mount_doc_cache = not path_is_relative_to(
            child=Path(self._args.collection_doc_cache_path),
            parent=(cache_path),
        ) and not path_is_relative_to(
            child=Path(self._args.collection_doc_cache_path),
            parent=(Path(playbook_dir)),
        )

        if mount_doc_cache:
            container_volume_mounts.append(
                f"{self._args.collection_doc_cache_path}:{self._args.collection_doc_cache_path}:z",
            )

        for volume_mount in container_volume_mounts:
            self._logger.debug("Adding volume mount to container invocation: %s", volume_mount)

        if "container_volume_mounts" in kwargs:
            kwargs["container_volume_mounts"] += container_volume_mounts
        else:
            kwargs["container_volume_mounts"] = container_volume_mounts

        return python_exec_path

    def _run_runner(self) -> None:
        """Use the runner subsystem to catalog collections."""
        if isinstance(self._args.set_environment_variable, dict):
            set_environment_variable = deepcopy(self._args.set_environment_variable)
        else:
            set_environment_variable = {}
        set_environment_variable["ANSIBLE_NOCOLOR"] = "True"

        if self._args.execution_environment:
            self._setup_ee_pythonpath(set_environment_variable)

        kwargs = {
            "container_engine": self._args.container_engine,
            "execution_environment_image": self._args.execution_environment_image,
            "execution_environment": self._args.execution_environment,
            "navigator_mode": "interactive",
            "pass_environment_variable": self._args.pass_environment_variable,
            "set_environment_variable": set_environment_variable,
            "private_data_dir": self._args.ansible_runner_artifact_dir,
            "rotate_artifacts": self._args.ansible_runner_rotate_artifacts_count,
            "timeout": self._args.ansible_runner_timeout,
        }

        if isinstance(self._args.playbook, str):
            playbook_dir = f"{Path(self._args.playbook).parent}"
        else:
            playbook_dir = str(Path.cwd())

        if isinstance(self._args.execution_environment_volume_mounts, list):
            kwargs["container_volume_mounts"] = self._args.execution_environment_volume_mounts

        if isinstance(self._args.container_options, list):
            kwargs["container_options"] = self._args.container_options

        kwargs["host_cwd"] = playbook_dir

        self._adjacent_collection_dir = f"{Path(playbook_dir) / 'collections'}"
        cache_path = self._args.internals.cache_path

        pass_through_arg = [
            f"{cache_path}/catalog_collections.py",
            "-a",
            self._adjacent_collection_dir,
            "-c",
            self._collection_cache_path,
        ]

        kwargs["cmdline"] = pass_through_arg

        if self._args.execution_environment:
            python_exec_path = self._setup_ee_volume_mounts(kwargs, playbook_dir, cache_path)
        else:
            self._logger.debug("running collections command locally")
            python_exec_path = sys.executable

        self._logger.debug(
            "Invoke runner with executable_cmd: %s and kwargs: %s",
            python_exec_path,
            kwargs,
        )
        _runner = Command(executable_cmd=python_exec_path, **kwargs)
        output, error, ret_code = _runner.run()

        if error:
            msg = f"Error while running catalog collection script: {error}"
            self._logger.error(msg)
        if ret_code:
            self._logger.error(output)
            self.notify_failed()
        if output:
            self._parse(output)

    def _extract_json_output(self, output: str) -> dict[str, Any] | None:
        """Extract and parse JSON from the collection cataloging output.

        Args:
            output: The raw output string that may contain warnings before JSON

        Returns:
            The parsed JSON dict, or None if parsing failed
        """
        try:
            if not output.startswith("{"):
                _warnings, json_str = output.split("{", 1)
                json_str = "{" + json_str
            else:
                json_str = output
            parsed = json.loads(json_str)
            self._logger.debug("json loading output succeeded")
        except ValueError as exc:
            self._logger.exception("Unable to extract collection json from stdout")
            self._logger.debug("error json loading output: '%s'", str(exc))
            self._logger.debug(output)
            return None
        return parsed

    def _get_dest_volume_mounts(self) -> tuple[str, ...]:
        """Build a tuple of destination volume mount paths for collection type detection.

        Returns:
            A tuple of destination paths derived from execution environment volume mounts
        """
        volume_mounts = self.app.args.execution_environment_volume_mounts
        if not isinstance(volume_mounts, list):
            return ()

        tmp_list = []
        for mount in volume_mounts:
            source_str, destination_str = mount.split(":")[0:2]
            if not Path(source_str).is_dir():
                continue
            dest_path = Path(destination_str)
            if dest_path.parent.name == "ansible_collections":
                continue
            if dest_path.parent.parent.name == "ansible_collections":
                continue
            if dest_path.name != "ansible_collections":
                dest_path /= "ansible_collections"
            tmp_list.append(str(dest_path))
        return tuple(tmp_list)

    def _annotate_collection(
        self,
        collection: dict[str, Any],
        dest_volume_mounts: tuple[str, ...],
    ) -> None:
        """Annotate a single collection with display metadata.

        Adds __name, __version, __shadowed, and (for EE runs) __type fields.

        Args:
            collection: The collection dict to annotate in place
            dest_volume_mounts: Destination volume mount paths for type detection
        """
        collection["__name"] = collection["known_as"]
        collection["__version"] = collection["collection_info"].get("version", "missing")
        collection["__shadowed"] = bool(collection["hidden_by"])
        if not self._args.execution_environment:
            return

        if collection["path"].startswith(dest_volume_mounts) or collection["path"].startswith(
            self._adjacent_collection_dir,
        ):
            collection["__type"] = "bind_mount"
        elif collection["path"].startswith(f"{Path(self._adjacent_collection_dir).parent}"):
            collection["__type"] = "bind_mount"
            error = (
                f"{collection['known_as']} was mounted and catalogued in the"
                " execution environment but was outside the adjacent 'collections'"
                " directory. This may cause issues outside the local development"
                " environment."
            )
            self._logger.error(error)
        else:
            collection["__type"] = "contained"

    def _parse(self, output: str) -> None:
        """Load and process the ``json`` output from the collection cataloging process.

        Args:
            output: The output from the collection cataloging process

        Returns:
            Nothing
        """
        parsed = self._extract_json_output(output)
        if parsed is None:
            return

        for error in parsed["errors"]:
            self._logger.error("%s %s", error["path"], error["error"])

        self._collections = sorted(
            parsed["collections"].values(),
            key=lambda i: i["known_as"],
        )

        dest_volume_mounts = self._get_dest_volume_mounts()

        for collection in self._collections:
            self._annotate_collection(collection, dest_volume_mounts)

        self._stats = parsed["stats"]

        if parsed.get("messages"):
            for msg in parsed["messages"]:
                self._logger.info("[catalog_collections]: %s", msg)

        self._logger.debug("catalog collections scan path: %s", parsed["collection_scan_paths"])
        self._collection_scanned_paths = parsed["collection_scan_paths"].split(":")
        for stat, value in self._stats.items():
            self._logger.debug("%s: %s", stat, value)

        if not parsed["collections"]:
            env = "execution" if self._args.execution_environment else "local"
            error = f"No collections found in {env} environment, searched in "
            error += parsed["collection_scan_paths"]
            self._logger.warning(error)

    def _extract_single_plugin_details(
        self,
        plugin_checksum: str,
        plugin_info: dict[str, Any],
        selected_collection: dict[str, Any],
    ) -> dict[str, Any]:
        """Extract details for a single plugin from the cache.

        Args:
            plugin_checksum: The checksum key for the plugin in the cache
            plugin_info: The plugin info dict containing type and path
            selected_collection: The parent collection data

        Returns:
            A dict with path, full_name, and optionally short_description
        """
        plugin_type = plugin_info.get("type")
        plugin_json = self._collection_cache[plugin_checksum]
        loaded = json.loads(plugin_json)

        plugin = loaded.get("plugin")
        plugin_docs: dict[str, Any] = {}
        plugin_path = Path(selected_collection.get("path", "")) / Path(
            plugin_info.get("path", ""),
        )
        plugin_docs["path"] = str(plugin_path)
        if plugin and plugin["doc"] is not None:
            try:
                if "name" in plugin["doc"]:
                    short_name = plugin["doc"]["name"]
                else:
                    short_name = plugin["doc"][plugin_type]
            except (KeyError, TypeError) as exc:
                short_name = None
                self._logger.exception(
                    "Error getting short name from plugin doc %s",
                    plugin_docs["path"],
                )
                self._logger.debug("error was %s", str(exc))
                self._logger.debug(plugin)
            if short_name is None:
                plugin_docs["full_name"] = selected_collection["known_as"]
            else:
                plugin_docs["full_name"] = f"{selected_collection['known_as']}.{short_name}"

            if "short_description" in plugin["doc"]:
                plugin_docs["short_description"] = plugin["doc"]["short_description"]

        return plugin_docs

    def _get_collection_plugins_details(
        self,
        selected_collection: dict[str, Any],
    ) -> dict[str, Any]:
        """Get plugin details for the given collection.

        Args:
            selected_collection: The selected collection

        Returns:
            The plugin details like full-name, type and short
            description.
        """
        plugins_details: dict[str, Any] = {}

        for plugin_checksum, plugin_info in selected_collection["plugin_checksums"].items():
            plugin_type = plugin_info.get("type")
            if plugin_type not in plugins_details:
                plugins_details[plugin_type] = []

            plugin_docs = self._extract_single_plugin_details(
                plugin_checksum,
                plugin_info,
                selected_collection,
            )
            plugins_details[plugin_type].append(plugin_docs)

        return plugins_details

    @staticmethod
    def _filter_roles_for_stdout(
        roles: list[dict[str, Any]],
        exclude_keys: list[str],
    ) -> list[dict[str, Any]]:
        """Filter excluded keys from role dicts for stdout display.

        Args:
            roles: The list of role dicts to filter
            exclude_keys: Keys to exclude from each role dict

        Returns:
            A list of role dicts with excluded keys removed
        """
        filtered = []
        for role in roles:
            updated_role_info = {k: v for k, v in role.items() if k not in exclude_keys}
            filtered.append(updated_role_info)
        return filtered

    def _format_collection_for_stdout(
        self,
        collection: dict[str, Any],
        collection_exclude_keys: list[str],
        roles_exclude_keys: list[str],
    ) -> dict[str, Any]:
        """Format a single collection for stdout display.

        Args:
            collection: The collection dict to format
            collection_exclude_keys: Top-level keys to exclude
            roles_exclude_keys: Keys to exclude from role entries

        Returns:
            The formatted collection dict with plugins details attached
        """
        plugins_details = self._get_collection_plugins_details(collection)

        collection_stdout: dict[str, Any] = {}
        for info_name, info_value in collection.items():
            info_name = remove_dbl_un(info_name)
            if info_name in collection_exclude_keys:
                continue

            if info_name == "roles":
                collection_stdout["roles"] = self._filter_roles_for_stdout(
                    info_value,
                    roles_exclude_keys,
                )
            else:
                collection_stdout[info_name] = info_value

        collection_stdout["plugins"] = plugins_details
        return collection_stdout

    def _parse_collection_info_stdout(self) -> dict[str, Any]:
        """Parse collection information from catalog collection cache.

        Returns:
            The collection information to be displayed on stdout
        """
        collection_exclude_keys = [
            "file_manifest_file",
            "format",
            "meta_source",
            "plugin_checksums",
            "runtime",
        ]
        roles_exclude_keys = ["readme"]

        self._collection_cache.open_()
        collections_info: dict[str, Any] = {
            "collections": [
                self._format_collection_for_stdout(
                    collection,
                    collection_exclude_keys,
                    roles_exclude_keys,
                )
                for collection in self._collections
            ],
        }
        self._collection_cache.close()

        return collections_info



"""Get one line of text input."""

from __future__ import annotations

import curses

from curses import ascii as curses_ascii
from curses.textpad import Textbox
from typing import TYPE_CHECKING

from .curses_defs import CursesLine
from .curses_defs import CursesLinePart
from .curses_window import CursesWindow
from .sentinels import unknown


if TYPE_CHECKING:
    from ansible_navigator.action_runner import Window
    from ansible_navigator.ui_framework.field_text import FieldText
    from ansible_navigator.ui_framework.ui_config import UIConfig


class FormHandlerText(CursesWindow, Textbox):
    """Get one line of text input."""

    def __init__(self, screen: Window, ui_config: UIConfig) -> None:
        """Initialize the handler for a form text field.

        Args:
            screen: A curses window
            ui_config: The current user interface configuration
        """
        super().__init__(ui_config=ui_config)
        self.input_line_cache: list[str] = []
        self.input_line_pointer = 0
        self._arrowing = False
        self.insert_mode = False
        self.stripspaces = True
        self._screen = screen

    def _paint_from_line_cache(
        self,
    ) -> None:
        """Put the text from the line cache in the text input window."""
        line_part = CursesLinePart(0, self.input_line_cache[self.input_line_pointer], 0, 0)
        self._add_line(self.win, 0, CursesLine((line_part,)))
        self.win.clrtoeol()

    def _adjust_line_pointer(self, amount: int) -> None:
        """Increase the current position in the line cache.

        Args:
            amount: The number to increase by
        """
        self.input_line_pointer = (self.input_line_pointer + amount) % len(self.input_line_cache)

    def _handle_arrow_down(self) -> int:
        """Handle arrow down / Ctrl-N key input for line cache navigation.

        Returns:
            1 to indicate the window should be repainted
        """
        if self.input_line_cache:
            if not self._arrowing:
                self.input_line_pointer = 0
                self._arrowing = True
            else:
                self._adjust_line_pointer(1)
                self._paint_from_line_cache()
        return 1

    def _handle_arrow_up(self) -> int:
        """Handle arrow up / Ctrl-P key input for line cache navigation.

        Returns:
            1 to indicate the window should be repainted
        """
        if self.input_line_cache:
            if not self._arrowing:
                self.input_line_pointer = len(self.input_line_cache) - 1
                self._arrowing = True
            else:
                self._adjust_line_pointer(-1)
                self._paint_from_line_cache()
        return 1

    def _do_command(self, char: int) -> int:
        """Handle the input from the user.

        Args:
            char: The character input from the user

        Returns:
            0 if the window needs to be resized 1 if the window should
            be repainted -1 if the window should be closed
        """
        # in the case the term returns 127 instead of 263
        if char == curses_ascii.DEL:
            char = curses.KEY_BACKSPACE

        if char == curses.KEY_IC:
            self.insert_mode = not self.insert_mode
            ret = 1
        elif char == curses_ascii.ESC:
            ret = -1
        elif char == curses.KEY_RESIZE:
            ret = 0
        elif char in (curses_ascii.SO, curses.KEY_DOWN):
            ret = self._handle_arrow_down()
        elif char in (curses_ascii.DLE, curses.KEY_UP):
            ret = self._handle_arrow_up()
        elif char == curses_ascii.TAB:
            ret = 0
        else:
            ret = bool(self.do_command(char))  # type: ignore[func-returns-value]
            self._arrowing = False
        return ret

    def handle(self, idx: int, form_fields: list[FieldText]) -> tuple[str, int]:
        """Edit in the widget window and collect the results.

        Args:
            idx: Index to retrieve specific field
            form_fields: List of fields

        Returns:
            Results from line and char edit
        """
        form_field = form_fields[idx]

        if form_field.response is not unknown:
            line_part = CursesLinePart(0, str(form_field.response), 0, 0)
            self._add_line(self.win, 0, CursesLine((line_part,)))
        else:
            self.win.move(0, 0)

        self._curs_set(1)
        while True:
            char = self.win.getch()

            if not char:
                continue
            cmd_res = self._do_command(char)
            if cmd_res == 0:
                break
            if cmd_res == -1:
                return "", char
            self.win.refresh()
        self._curs_set(0)
        line = self.gather().strip()
        if line:
            self.input_line_cache.append(line)
        return line, char



"""``:quit`` command implementation."""

import logging

from ansible_navigator.app_public import AppPublic
from ansible_navigator.configuration_subsystem.definitions import ApplicationConfiguration
from ansible_navigator.ui_framework import Interaction

from . import _actions as actions


@actions.register
class Action:
    """``:quit`` command implementation."""

    KEGEX = r"q(?:uit)?(?P<exclamation>!)?$"

    def __init__(self, args: ApplicationConfiguration) -> None:
        """Initialize the ``:quit`` action.

        Args:
            args: The current settings for the application
        """
        self._args = args
        self._logger = logging.getLogger(__name__)

    def run(self, interaction: Interaction, app: AppPublic) -> Interaction:
        """Handle a request to quit the application from the user interface.

        The application exit is ultimately handled by
        :class:`~ansible_navigator.action_runner.ActionRunner` so return immediately,
        passing the ``:quit`` request backwards through the stack of actions that are on the stack.

        Args:
            interaction: The interaction from the user
            app: The app instance

        Returns:
            The pending
            :class:`~ansible_navigator.ui_framework.ui.Interaction`
        """
        self._logger.debug("quit was requested as: %s", interaction.action.value)
        return interaction



"""Get one line of text input."""

from __future__ import annotations

import curses

from curses import ascii as curses_ascii
from typing import TYPE_CHECKING
from typing import Any

from .curses_defs import CursesLine
from .curses_defs import CursesLinePart
from .curses_window import CursesWindow


if TYPE_CHECKING:
    from ansible_navigator.action_runner import Window
    from ansible_navigator.ui_framework.field_text import FieldText
    from ansible_navigator.ui_framework.ui_config import UIConfig


class FormHandlerOptions(CursesWindow):
    """Handle form checkbox field."""

    def __init__(self, screen: Window, ui_config: UIConfig) -> None:
        """Initialize the handler for either form checkboxes or radio buttons.

        Args:
            screen: A curses window
            ui_config: The current user interface configuration
        """
        super().__init__(ui_config=ui_config)
        self._screen = screen

    def populate(self, form_field: FieldText, active: int) -> None:
        """Populate the window with the checkboxes.

        Args:
            form_field: Field from a form
            active: Track active checkbox/option
        """
        options = getattr(form_field, "options", [])
        for idx, option in enumerate(options):
            option_code = option.ansi_code(form_field)
            color = 8 if option.disabled else 0
            decoration = curses.A_STANDOUT if idx == active else 0
            clp_option_code = CursesLinePart(0, option_code, color, 0)
            if self._ui_config.color is False:
                text = (
                    f"[{option.text.capitalize()}]" if idx == active else option.text + "  "
                )  # clear the [], for else
            else:
                text = option.text
            clp_text = CursesLinePart(len(option_code) + 1, text, color, decoration)
            self._add_line(
                window=self.win,
                lineno=idx,
                line=CursesLine((clp_option_code, clp_text)),
            )

    def _handle_key_input(
        self,
        char: int,
        form_field: Any,
        active: int,
    ) -> tuple[bool, int]:
        """Process a single key input for checkbox/radio fields.

        Args:
            char: The character input from the user
            form_field: Field from a form
            active: Track active checkbox/option

        Returns:
            Tuple of (should_break, new_active)
        """
        if char in (curses_ascii.SO, curses.KEY_DOWN):
            return False, active + 1

        if char in (curses_ascii.DLE, curses.KEY_UP):
            return False, active - 1

        if char == curses.KEY_RESIZE:
            return True, active

        if char == curses_ascii.TAB:
            if active == len(form_field.options) - 1:
                return True, active
            return False, active + 1

        if char == curses_ascii.SP:
            if not form_field.options[active].disabled:
                form_field.options[active].checked = not form_field.options[active].checked
                if form_field.__class__.__name__ == "FieldRadio":
                    for ffo_idx, option in enumerate(form_field.options):
                        if ffo_idx != active:
                            option.checked = False
            return False, active

        if char in [curses_ascii.NL, curses_ascii.CR]:
            return True, active

        return False, active

    def handle(self, idx: int, form_fields: list[FieldText]) -> tuple[FieldText, int]:
        """Handle the check box field.

        Args:
            form_fields: List of fields
            idx: Index to retrieve specific field

        Returns:
            Field from form and characters

        Raises:
            RuntimeError: if there is a runtime error.
        """
        form_field = form_fields[idx]
        active = 0

        if not hasattr(form_field, "options"):
            raise RuntimeError
        while True:
            active = active % len(form_field.options)
            self.populate(form_field, active)

            char = self.win.getch()

            should_break, active = self._handle_key_input(char, form_field, active)
            if should_break:
                break

        return form_field, char



"""Definitions of UI content objects."""

from __future__ import annotations

from collections.abc import ItemsView
from collections.abc import Sequence
from dataclasses import asdict
from dataclasses import dataclass
from enum import Enum
from typing import TYPE_CHECKING
from typing import Any
from typing import Generic
from typing import TypeVar


if TYPE_CHECKING:
    from .utils.compatibility import TypeAlias


class ContentView(Enum):
    """The content view."""

    FULL = "full"
    NORMAL = "normal"


T = TypeVar("T")
DictType: TypeAlias = dict[str, T]


class SerializationFormat(Enum):
    """The serialization format."""

    YAML = "YAML"
    JSON = "JSON"


@dataclass
class ContentBase(Generic[T]):
    r"""The base class for all content dataclasses presented in the UI.

    It should be noted, that while the return type is defined as ``T``
    for the serialization functions below, mypy will not catch in incorrect
    definition of ``T`` at this time.  This is because of how ``asdict()``
    is typed:

    @overload
    def asdict(obj: Any) -> dict[str, Any]: ...
    @overload
    def asdict(obj: Any, \*, dict_factory: Callable[[list[tuple[str, Any]]], _T]) -> _T: ...

    Which result in mypy believing the outcome of asdict is dict[str, Any] and letting it silently
    pass through an incorrect ``T``. ``Mypy`` identifies this as a known issue:
    https://mypy.readthedocs.io/en/stable/additional_features.html#caveats-known-issues
    """

    def asdict(
        self,
        content_view: ContentView,
        serialization_format: SerializationFormat,
    ) -> DictType[Any]:
        """Convert thy self into a dictionary.

        Args:
            content_view: The content view
            serialization_format: The serialization format

        Returns:
            A dictionary created from self
        """
        converter_map = {
            (ContentView.FULL, SerializationFormat.JSON): self.serialize_json_full,
            (ContentView.FULL, SerializationFormat.YAML): self.serialize_yaml_full,
            (ContentView.NORMAL, SerializationFormat.JSON): self.serialize_json_normal,
            (ContentView.NORMAL, SerializationFormat.YAML): self.serialize_yaml_normal,
        }

        try:
            dump_self_as_dict = converter_map[content_view, serialization_format]
        except KeyError:
            return asdict(self)
        return dump_self_as_dict()

    def serialize_json_full(self) -> DictType[Any]:
        """Provide dictionary for ``JSON`` with all attributes.

        Returns:
            A dictionary created from self
        """
        return asdict(self)

    def serialize_json_normal(self) -> DictType[Any]:
        """Provide dictionary for ``JSON`` with curated attributes.

        Returns:
            A dictionary created from self
        """
        return asdict(self)

    def serialize_yaml_full(self) -> DictType[Any]:
        """Provide dictionary for ``YAML`` with all attributes.

        Returns:
            A dictionary created from self
        """
        return asdict(self)

    def serialize_yaml_normal(self) -> DictType[Any]:
        """Provide dictionary for ``JSON`` with curated attributes.

        Returns:
            A dictionary created from self
        """
        return asdict(self)

    def get(self, attribute: str) -> Any:
        """Allow this dataclass to be treated like a dictionary.

        This is a work around until the UI fully supports dataclasses
        at which time this can be removed.

        Default is intentionally not implemented as a safeguard to enure
        this is not more work than necessary to remove in the future
        and will only return attributes in existence.

        Args:
            attribute: The attribute to get

        Returns:
            The gotten attribute
        """
        return getattr(self, attribute)

    def items(self) -> ItemsView[str, Any]:
        """Allow this dataclass to be treated like a dictionary.

        This is a work around until the UI fully supports dataclasses
        at which time this can be removed.

        Returns:
            Tuples of attribute value pairs
        """
        return asdict(self).items()


ContentTypeSingle = bool | float | int | str | dict[str, Any] | ContentBase[Any]
ContentTypeSequence = list[Any] | Sequence[ContentBase[Any]]
ContentType = ContentTypeSingle | ContentTypeSequence


@dataclass(frozen=True)
class CFormat:
    """A single instance for a content format.

    Attributes:
        scope: The scope, used for tokenization.
        file_extension: The file extension, with a dot.
        serialization: If needed the serialization format.
    """

    scope: str
    file_extension: str
    serialization: SerializationFormat | None


class ContentFormat(Enum):
    """All content formats."""

    ANSI = CFormat(scope="source.ansi", file_extension=".ansi", serialization=None)
    JSON = CFormat(
        scope="source.json",
        file_extension=".json",
        serialization=SerializationFormat.JSON,
    )
    LOG = CFormat(scope="text.log", file_extension=".log", serialization=None)
    MARKDOWN = CFormat(scope="text.html.markdown", file_extension=".md", serialization=None)
    TXT = CFormat(scope="source.txt", file_extension=".txt", serialization=None)
    YAML = CFormat(
        scope="source.yaml",
        file_extension=".yml",
        serialization=SerializationFormat.YAML,
    )
    # YAML as string, already serialized
    YAML_TXT = CFormat(
        scope="source.yaml",
        file_extension=".yml",
        serialization=None,
    )



"""Utilities related to the merging of dictionaries."""

from __future__ import annotations

from typing import Any


class DictMergeError(Exception):
    """Custom exception for a dict merge error."""


Mergeable = bool | dict[Any, Any] | list[Any] | str | None


def _merge_dicts(
    left: dict[Any, Any],
    right: dict[Any, Any],
) -> dict[Any, Any]:
    """Merge right dict into left dict recursively.

    Args:
        left: Dict to be merged into.
        right: Dict to be merged from.

    Returns:
        The merged dict.
    """
    for key, value in right.items():
        if key in left:
            left[key] = in_place_list_replace(left[key], value)
        else:
            left[key] = value
    return left


def in_place_list_replace(left: Mergeable, right: Mergeable) -> Mergeable:
    """Merge right into left in place and returns merged result.

    If left is a primitive, right will be used
    If left key is a list or tuple right will be used.
    If left and right have a common key, they will be dict merged
    If left is missing a key from right, the key from right will be added

    Args:
        left: Dict to be merged into
        right: Dict to be merged from

    Raises:
        DictMergeError: If merging cannot be done

    Returns:
        Merged dict
    """
    key = None
    try:
        if left is None or isinstance(left, (str | int | float | bool, tuple | list)):
            left = right
        elif isinstance(left, dict):
            if isinstance(right, dict):
                left = _merge_dicts(left, right)
            else:
                msg = f"Cannot merge non-dict '{right}' into dict '{left}'"
                raise DictMergeError(msg)
        else:
            msg = f"Not implemented '{right}' into '{left}'"
            raise DictMergeError(msg)
    except TypeError as exc:
        msg = f"TypeError '{exc}' in key '{key}' when merging '{right}' into '{left}'"
        raise DictMergeError(msg) from exc
    return left



# cspell:ignore envvars
"""Herein lies the base class for all interaction with ansible-runner.

Attributes common to all ansible-runner calls are defined within the base class.
"""

from __future__ import annotations

import logging
import os
import shutil
import sys
import tempfile

from pathlib import Path
from typing import TYPE_CHECKING
from typing import Any


if TYPE_CHECKING:
    from ansible_runner import Runner


class Base:
    """Base class for ansible-runner calls."""

    # pylint: disable=too-many-arguments
    # pylint: disable=too-many-instance-attributes
    def __init__(
        self,
        private_data_dir: str | None = None,
        container_engine: str | None = None,
        execution_environment: bool | None = False,
        execution_environment_image: str | None = None,
        navigator_mode: str | None = None,
        container_volume_mounts: list[str] | None = None,
        container_options: list[str] | None = None,
        container_workdir: str | None = None,
        set_environment_variable: dict[str, str] | None = None,
        pass_environment_variable: list[str] | None = None,
        host_cwd: str | None = None,
        rotate_artifacts: int | None = None,
        timeout: int | None = None,
    ) -> None:
        """Handle the common argument for the ansible-runner interface class.

        Args:
            container_engine: Container engine used to isolate execution
            container_options: List of container options to pass to
                execution engine
            execution_environment: Boolean argument that enable
                execution environment support
            execution_environment_image: Container image to use when
                running an command
            navigator_mode: Valid value is either ``stdout`` or
                ``interactive``. If value is set to ``stdout`` passed
                the ``stdin`` of current running process is passed to
                ``ansible-runner`` which enables receiving command line
                prompts after the executing the command. If value is set
                to ``interactive`` the ``ansible-navigator` will run
                using text user interface (TUI).
            container_volume_mounts: List of bind mounts in the form
                ``host_dir:/container_dir:labels``
            container_workdir: The working directory within the
                container
            host_cwd: The current local working directory. If value of
                execution_environment is set to True this path will be
                volume mounted within the execution environment.
            set_environment_variable: Dict of user requested environment
                variables to set
            private_data_dir: The directory containing all runner
                metadata needed to invoke the runner module. Output
                artifacts will also be stored here for later
                consumption.
            pass_environment_variable: List of user requested
                environment variables to pass
            rotate_artifacts: Keep at most n artifact directories
            timeout: The timeout value in seconds that will be passed to
                either ``pexpect`` of ``subprocess`` invocation (based
                on ``runner_mode`` selected) while executing command. It
                the timeout is triggered it will force cancel the
                execution.
        """
        self._logger = logging.getLogger(__name__)

        self._set_private_data_directory(private_data_dir)
        self._ce = container_engine
        self._ee = execution_environment
        self._eei = execution_environment_image
        self._navigator_mode = navigator_mode
        self._set_environment_variable: dict[str, Any] = (
            set_environment_variable if isinstance(set_environment_variable, dict) else {}
        )
        self._pass_environment_variable: list[str] = (
            pass_environment_variable if isinstance(pass_environment_variable, list) else []
        )
        self._host_cwd = host_cwd
        self._rotate_artifacts = rotate_artifacts if isinstance(rotate_artifacts, int) else None
        self._timeout = timeout if isinstance(timeout, int) else None
        self.ansible_runner_instance: Runner
        self.cancelled: bool = False
        self.finished: bool = False
        self.status: str | None = None
        self._runner_args: dict[str, Any] = {}

        container_options = container_options or []

        # when the ce is podman, set the container user to root unless already specified
        if self._ce == "podman":
            has_user_flag = any(
                opt.startswith(("--user", "-u=", "-u ")) or opt == "-u" for opt in container_options
            )
            if not has_user_flag:
                container_options.append("--user=root")

        # Fix SSH agent socket when running Docker on macOS.
        if sys.platform == "darwin" and self._ce == "docker":
            ssh_agent_socket_opts = [
                # Docker exposes this proxy socket for SSH agent on macOS.
                "--volume=/run/host-services/ssh-auth.sock:/run/host-services/ssh-auth.sock:ro",
                "--env=SSH_AUTH_SOCK=/run/host-services/ssh-auth.sock",
            ]

            # Prepend to existing container options to allow overriding this fix.
            container_options = ssh_agent_socket_opts + container_options

        # On macOS Tahoe (>26.3), the launchd SSH agent socket moved from
        # /private/tmp to /var/run. Since /var/run is a symlink to /private/var/run,
        # container runtimes require the resolved real path when building volume mount
        # arguments. ansible-runner normalises mount paths with os.path.abspath(), which
        # does not resolve symlinks, causing the container engine to fail with:
        #   Error: statfs /var/run/com.apple.launchd.xxxxxxxxx: no such file or directory
        # Resolving SSH_AUTH_SOCK to its real path here — before ansible-runner reads it
        # from os.environ in _handle_automounts() — ensures the correct path is passed to
        # the container engine regardless of which CE is in use.
        if sys.platform == "darwin" and "SSH_AUTH_SOCK" in os.environ:
            _real_sock = os.path.realpath(os.environ["SSH_AUTH_SOCK"])
            if _real_sock != os.environ["SSH_AUTH_SOCK"]:
                self._logger.debug(
                    "macOS SSH_AUTH_SOCK symlink resolved: %s -> %s",
                    os.environ["SSH_AUTH_SOCK"],
                    _real_sock,
                )
                os.environ["SSH_AUTH_SOCK"] = _real_sock

        if self._ee:
            self._runner_args.update(
                {
                    "container_image": self._eei,
                    "process_isolation_executable": self._ce,
                    "process_isolation": True,
                    "container_volume_mounts": container_volume_mounts,
                    "container_options": container_options,
                    "container_workdir": container_workdir,
                },
            )
        self._runner_args.update(
            {
                "private_data_dir": self._private_data_dir,
                "json_mode": True,
                "quiet": True,
                "cancel_callback": self.runner_cancelled_callback,
                "finished_callback": self.runner_finished_callback,
                "timeout": self._timeout,
            },
        )
        if self._rotate_artifacts is not None:
            self._runner_args["rotate_artifacts"] = self._rotate_artifacts

        self._add_env_vars_to_args()

        if self._host_cwd:
            # ensure the CWD ends with a trailing slash
            host_cwd = f"{Path(self._host_cwd) / ''}"
            self._runner_args["host_cwd"] = host_cwd

        if self._navigator_mode == "stdout":
            self._runner_args.update(
                {
                    "input_fd": sys.stdin,
                    "output_fd": sys.stdout,
                    "error_fd": sys.stderr,
                },
            )

    def __del__(self) -> None:
        """Drop the private_data_dir if it is temporary."""
        if (
            self._private_data_dir_is_tmp
            and self._private_data_dir
            and Path(self._private_data_dir).exists()
        ):
            self._logger.debug(
                "delete temporary ansible-runner private_data_dir at path %s",
                self._private_data_dir,
            )
            shutil.rmtree(self._private_data_dir, ignore_errors=True)

    @staticmethod
    def _generate_tmp_directory() -> str:
        """Generate a tmp directory for artifacts.

        Returns:
            Temporary directory path
        """
        return tempfile.mkdtemp(prefix="ansible-navigator_")

    def _set_private_data_directory(self, provided: str | None) -> None:
        """Set the private data directory to the provided, the username, or a uuid.

        Args:
            provided: Data directory path
        """
        if isinstance(provided, str):
            if os.access(provided, os.W_OK | os.R_OK | os.X_OK):
                private_data_directory = provided
                source = "user provided"
            else:
                self._logger.debug("Provided private data dir `%s` was not user writable", provided)
                private_data_directory = self._generate_tmp_directory()
                source = "user provided, but changed to tmp location due to permissions"
        else:
            private_data_directory = self._generate_tmp_directory()
            source = "not user provided, used tmp location"

        self._private_data_dir_is_tmp = source != "user provided"
        self._private_data_dir = private_data_directory
        self._logger.debug("private data dir %s: %s", source, self._private_data_dir)

    def runner_cancelled_callback(self) -> bool:
        """Check by runner to see if it should cancel.

        Returns:
            Boolean cancelled value
        """
        return self.cancelled

    def runner_finished_callback(self, runner: Runner) -> None:
        """Call when runner finishes.

        Args:
            runner: A runner instance
        """
        self.status = runner.status
        self.finished = True

    def _add_env_vars_to_args(self) -> None:
        """Add environment variables to runner args."""
        self._runner_args["envvars"] = {
            k: v for k, v in os.environ.items() if k.startswith("ANSIBLE_")
        }
        self._runner_args["envvars"].update(self._set_environment_variable)
        for env_var in self._pass_environment_variable:
            value = os.environ.get(env_var)
            if value is None:
                self._logger.warning(
                    "Pass through environment variable `%s` not currently set, discarded",
                    env_var,
                )
            else:
                self._runner_args["envvars"][env_var] = value



"""Definitions for image inspection."""

from __future__ import annotations

import json
import re

from collections.abc import Iterable
from typing import Any

from ansible_navigator.command_runner import Command
from ansible_navigator.command_runner import CommandRunner
from ansible_navigator.utils.functions import pascal_to_snake


class ImagesInspect:
    """Functionality for inspecting container images."""

    def __init__(self, container_engine: str, ids: list[str]) -> None:
        """Initialize the container image inspector.

        Args:
            container_engine: The name of the container engine to use
            ids: The ids of the container images to inspect
        """
        self._container_engine = container_engine
        self._image_ids = ids

    @property
    def commands(self) -> list[Command]:
        """Generate image inspection commands.

        Returns:
            List of image inspection command objects
        """
        return [
            Command(
                identity=image_id,
                command=f"{self._container_engine} inspect {image_id}",
                post_process=self.parse,
            )
            for image_id in self._image_ids
        ]

    @staticmethod
    def parse(command: Command) -> None:
        """Parse the image inspection command output.

        Args:
            command: Image inspection command object
        """
        obj = json.loads(command.stdout)
        snake = pascal_to_snake(obj[0])
        command.details = snake


class ImagesList:
    """Functionality for listing container images."""

    def __init__(self, container_engine: str) -> None:
        """Initialize the container image lister.

        Args:
            container_engine: The name of the container engine to use
        """
        self._container_engine = container_engine

    @property
    def commands(self) -> list[Command]:
        """Generate image lister commands.

        Returns:
            List of the image lister commands
        """
        return [
            Command(
                identity="images",
                command=f"{self._container_engine} images",
                post_process=self.parse,
            ),
        ]

    @staticmethod
    def parse(command: Command) -> None:
        """Parse the image lister command output.

        Args:
            command: Image lister command object
        """
        if command.stdout:
            images = command.stdout.splitlines()
            re_2omo = re.compile(r"\s{2,}")
            headers = [key.lower().replace(" ", "_") for key in re_2omo.split(images.pop(0))]
            local_images = [
                dict(zip(headers, re_2omo.split(line), strict=False)) for line in images
            ]
            valid_images = [image for image in local_images if image.get("tag") != "<none>"]
            command.details = valid_images


def inspect_all(container_engine: str) -> tuple[list[dict[str, Any]], str]:
    """Run inspect against all images in the list.

    Args:
        container_engine: Name of the container engine

    Returns:
        List of all image values and stderr, if applicable

    Raises:
        TypeError: If there is a typing error
    """
    cmd_runner = CommandRunner()
    images_list_class = ImagesList(container_engine=container_engine)
    result = cmd_runner.run_single_process(commands=images_list_class.commands)
    images_list = result[0]
    if images_list.errors:
        return [], images_list.errors
    if images_list.stderr and not images_list.details:
        return [], images_list.stderr
    if not isinstance(images_list.details, Iterable):
        raise TypeError
    images = {image["image_id"]: image for image in images_list.details}
    image_ids = [image["image_id"] for image in images.values()]
    images_inspect_class = ImagesInspect(container_engine=container_engine, ids=image_ids)
    inspects = cmd_runner.run_single_process(commands=images_inspect_class.commands)
    for inspect in inspects:
        images[inspect.identity]["inspect"] = {"details": inspect.details, "errors": inspect.errors}
    return list(images.values()), images_list.stderr



"""Initialization file for the ui_framework."""

from .curses_defs import CursesLine
from .curses_defs import CursesLinePart
from .curses_defs import CursesLines
from .form_utils import dict_to_form
from .form_utils import error_notification
from .form_utils import form_to_dict
from .form_utils import nonblocking_notification
from .form_utils import success_notification
from .form_utils import warning_notification
from .ui import Action
from .ui import Content
from .ui import Interaction
from .ui import Menu
from .ui import UserInterface
from .ui_config import UIConfig
from .ui_constants import Color
from .ui_constants import Decoration


__all__ = (
    "Action",
    "Color",
    "Content",
    "CursesLine",
    "CursesLinePart",
    "CursesLines",
    "Decoration",
    "Interaction",
    "Menu",
    "UIConfig",
    "UserInterface",
    "dict_to_form",
    "error_notification",
    "form_to_dict",
    "nonblocking_notification",
    "success_notification",
    "warning_notification",
)



"""Initialization helpers that are used early in application initialization.

These helpers are specific to ansible_navigator.
"""

from __future__ import annotations

import logging
import os
import sys

from pathlib import Path
from typing import TYPE_CHECKING
from typing import NoReturn

from ._version_doc_cache import __version_collection_doc_cache__ as VERSION_CDC
from .configuration_subsystem import Configurator
from .configuration_subsystem import Constants as C
from .diagnostics import DiagnosticsCollector
from .utils.definitions import ExitMessage
from .utils.definitions import ExitMessages
from .utils.definitions import ExitPrefix
from .utils.definitions import LogMessage
from .utils.functions import console_width
from .utils.functions import environment_variable_is_file_path
from .utils.functions import find_settings_file
from .utils.key_value_store import KeyValueStore


if TYPE_CHECKING:
    from .configuration_subsystem.definitions import ApplicationConfiguration


def error_and_exit_early(exit_messages: list[ExitMessage]) -> NoReturn:
    """Exit the application early.

    Args:
        exit_messages: List of all exit messages to be printed
    """
    color = "NO_COLOR" not in os.environ
    printable = ExitMessages(messages=exit_messages).to_strings(color=color, width=console_width())
    print("\n".join(printable), file=sys.stderr)
    sys.exit(1)


def find_config() -> tuple[list[LogMessage], list[ExitMessage], str | None, C]:
    """Find a configuration file, logging each step.

    If the config can't be found/loaded, use default settings.
    If it's found but empty or not well formed, bail out.

    Returns:
        All log messages and config path
    """
    messages: list[LogMessage] = []
    exit_messages: list[ExitMessage] = []
    config_path = None
    settings_source = C.NONE

    # Check if the configuration file path is set via an environment var
    cfg_env_var = "ANSIBLE_NAVIGATOR_CONFIG"
    new_messages, new_exit_messages, env_config_path = environment_variable_is_file_path(
        env_var=cfg_env_var,
        kind="config",
    )
    messages.extend(new_messages)
    exit_messages.extend(new_exit_messages)
    if exit_messages:
        return messages, exit_messages, config_path, settings_source

    # Find the setting file
    new_messages, new_exit_messages, found_config_path = find_settings_file()

    messages.extend(new_messages)
    exit_messages.extend(new_exit_messages)
    if exit_messages:
        return messages, exit_messages, config_path, settings_source

    # Pick the environment variable set first, followed by found, followed by leave as none
    if env_config_path is not None:
        config_path = env_config_path
        settings_source = C.ENVIRONMENT_VARIABLE
        message = f"Using settings file at {config_path} set by {cfg_env_var}"
        messages.append(LogMessage(level=logging.DEBUG, message=message))
    elif found_config_path is not None:
        config_path = str(found_config_path)
        settings_source = C.SEARCH_PATH
        message = f"Using settings file at {config_path} in search path"
        messages.append(LogMessage(level=logging.DEBUG, message=message))
    else:
        message = "No valid settings file found, using all default values for settings."
        messages.append(LogMessage(level=logging.DEBUG, message=message))
    return messages, exit_messages, config_path, settings_source


def get_and_check_collection_doc_cache(
    collection_doc_cache_path: str,
) -> tuple[list[LogMessage], list[ExitMessage], KeyValueStore | None]:
    """Ensure the collection doc cache has current application version as a safeguard.

    Always delete and rebuild if not.

    Args:
        collection_doc_cache_path: Path for collection documentation
            cache

    Returns:
        All messages and collection cache or None
    """
    messages: list[LogMessage] = []
    exit_messages: list[ExitMessage] = []
    message = f"Collection doc cache: 'path' is '{collection_doc_cache_path}'"
    messages.append(LogMessage(level=logging.DEBUG, message=message))

    path_errors = []
    doc_cache_dir = Path(collection_doc_cache_path).parent
    try:
        doc_cache_dir.mkdir(parents=True, exist_ok=True)
    except OSError as exc:
        path_errors.append(f"Problem making directory: {doc_cache_dir!s}")
        path_errors.append(f"Error was: {exc!s}")

    if not os.access(Path(collection_doc_cache_path).parent, os.W_OK):
        path_errors.append("Directory not writable")

    if not os.access(Path(collection_doc_cache_path).parent, os.R_OK):
        path_errors.append("Directory not readable")

    if path_errors:
        exit_msgs = ["Problem while building the collection doc cache."]
        exit_msgs.extend(path_errors)
        exit_messages.extend([ExitMessage(message=exit_msg) for exit_msg in exit_msgs])
        exit_msg = "Try again without '--cdcp' or try '--cdcp ~/collection_doc_cache.db"
        exit_messages.append(ExitMessage(message=exit_msg, prefix=ExitPrefix.HINT))
        return messages, exit_messages, None

    collection_cache: KeyValueStore = KeyValueStore(collection_doc_cache_path)
    cache_version = collection_cache.get("version", None)
    message = f"Collection doc cache: 'current version' is '{cache_version}'"
    messages.append(LogMessage(level=logging.DEBUG, message=message))
    if cache_version is None or cache_version != VERSION_CDC:
        message = "Collection doc cache: version was empty or incorrect, rebuilding"
        messages.append(LogMessage(level=logging.INFO, message=message))
        collection_cache.close()
        Path(collection_doc_cache_path).unlink()
        collection_cache = KeyValueStore(collection_doc_cache_path)
        collection_cache["version"] = VERSION_CDC
        cache_version = collection_cache["version"]
        message = f"Collection doc cache: 'current version' is '{cache_version}'"
        messages.append(LogMessage(level=logging.INFO, message=message))
    collection_cache.close()
    return messages, exit_messages, collection_cache


def _diagnose(
    args: ApplicationConfiguration,
    exit_messages: list[ExitMessage],
    messages: list[LogMessage],
    should_diagnose: bool,
) -> tuple[list[LogMessage], list[ExitMessage]]:
    """Direct to the diagnostic information writer or return.

    Args:
        args: Application configuration
        exit_messages: List of exit messages
        messages: List of log messages
        should_diagnose: Whether to show tech info

    Returns:
        All messages and exit messages
    """
    if should_diagnose:
        DiagnosticsCollector(args=args, messages=messages, exit_messages=exit_messages).run()
    else:
        return messages, exit_messages
    return ([], [])


def parse_and_update(
    params: list[str],
    args: ApplicationConfiguration,
    apply_previous_cli_entries: C | list[str] = C.NONE,
    attach_cdc: bool = False,
) -> tuple[list[LogMessage], list[ExitMessage]]:
    """Build a configuration.

    Return after the CDC is mounted, even if exit messages are generated, the CDC may still
    be needed. e.g. ``:collections --ee NotBool``.

    Args:
        params: A list of parameters e.g. ['-x', 'value']
        args: The application args
        apply_previous_cli_entries: Should previous params from the CLI
            be applied
        attach_cdc: Should the collection doc cache be attached to the
            args.internals

    Returns:
        Log and exit messages
    """
    messages: list[LogMessage] = []
    exit_messages: list[ExitMessage] = []
    should_diagnose = "--diagnostics" in params

    (
        new_messages,
        new_exit_messages,
        args.internals.settings_file_path,
        args.internals.settings_source,
    ) = find_config()
    messages.extend(new_messages)
    exit_messages.extend(new_exit_messages)
    if exit_messages:
        return _diagnose(
            args=args,
            exit_messages=exit_messages,
            messages=messages,
            should_diagnose=should_diagnose,
        )

    configurator = Configurator(
        params=params,
        application_configuration=args,
        apply_previous_cli_entries=apply_previous_cli_entries,
        skip_roll_back=should_diagnose,
    )

    new_messages, new_exit_messages = configurator.configure()
    messages.extend(new_messages)
    exit_messages.extend(new_exit_messages)

    if args.internals.collection_doc_cache is C.NOT_SET:
        mount_collection_cache = True
        message = "Collection doc cache not mounted"
        messages.append(LogMessage(level=logging.DEBUG, message=message))
    elif args.initial.collection_doc_cache_path != args.collection_doc_cache_path:
        mount_collection_cache = True
        message = "Collection doc cache path changed"
        messages.append(LogMessage(level=logging.DEBUG, message=message))
    else:
        mount_collection_cache = False

    if mount_collection_cache and isinstance(args.collection_doc_cache_path, str):
        new_messages, new_exit_messages, cache = get_and_check_collection_doc_cache(
            args.collection_doc_cache_path,
        )
        messages.extend(new_messages)
        exit_messages.extend(new_exit_messages)
        if cache is None:
            # There's nothing to be done here, it cannot be attached.
            return _diagnose(
                args=args,
                exit_messages=exit_messages,
                messages=messages,
                should_diagnose=should_diagnose,
            )
        if attach_cdc:
            args.internals.collection_doc_cache = cache
            message = "Collection doc cache attached to args.internals"
            messages.append(LogMessage(level=logging.DEBUG, message=message))
        else:
            message = "Collection doc cache not attached to args.internals"
            messages.append(LogMessage(level=logging.DEBUG, message=message))

    if exit_messages:
        return _diagnose(
            args=args,
            exit_messages=exit_messages,
            messages=messages,
            should_diagnose=should_diagnose,
        )

    for entry in args.entries:
        message = f"Running with {entry.name} as '{entry.value.current}'"
        message += f" ({type(entry.value.current).__name__}/{entry.value.source.value})"
        messages.append(LogMessage(level=logging.DEBUG, message=message))

    return _diagnose(
        args=args,
        exit_messages=exit_messages,
        messages=messages,
        should_diagnose=should_diagnose,
    )



"""Common form definitions."""

from __future__ import annotations

from enum import Enum

from .sentinels import Unknown


FieldValidationState = Unknown | bool
FieldValidationStates = list[FieldValidationState]


class FormType(Enum):
    """Indicates the form type."""

    FORM = "A full input form"
    NOTIFICATION = "A text box with a dismiss button"
    WORKING = "Used when something might take a bit, no buttons"



"""Commonly used definitions."""

from __future__ import annotations

from dataclasses import dataclass
from typing import NamedTuple
from typing import NewType


class CursesLinePart(NamedTuple):
    """One chunk of a line of text.

    Args:
        column: the column at which the text should start
        string: the text to be displayed
        color: An integer representing a color, not a
            curses.color_pair(n)
        decoration: A curses decoration
    """

    column: int
    string: str
    color: int
    decoration: int


CursesLine = NewType("CursesLine", tuple[CursesLinePart, ...])
"""One line of text ready for curses."""

CursesLines = NewType("CursesLines", tuple[CursesLine, ...])
"""One or more lines of text ready for curses."""


RgbTuple = tuple[int, int, int]


@dataclass
class SimpleLinePart:
    """Definition of one part of one line having a common color."""

    #: One group of characters sharing the same color
    chars: str
    #: The column where the characters start, the sum of all previous characters in the line
    column: int
    #: The color for these characters
    color: RgbTuple | None
    #: The style for these characters
    style: str | None



"""Produce a diagnostics report in json format."""

from __future__ import annotations

import sys
import traceback

from dataclasses import asdict
from dataclasses import dataclass
from datetime import datetime
from datetime import timezone
from importlib.util import find_spec
from pathlib import Path
from sys import stdout
from typing import TYPE_CHECKING
from typing import Any

from .command_runner import Command
from .command_runner import CommandRunner
from .configuration_subsystem import Constants
from .configuration_subsystem import to_effective
from .configuration_subsystem import to_sources
from .data import image_introspect
from .image_manager import introspector
from .utils import ansi
from .utils.compatibility import importlib_metadata
from .utils.functions import now_iso
from .utils.functions import shlex_join
from .utils.serialize import Loader
from .utils.serialize import write_diagnostics_json
from .utils.serialize import yaml


if TYPE_CHECKING:
    from collections.abc import Callable
    from collections.abc import Iterator

    from .configuration_subsystem.definitions import ApplicationConfiguration
    from .utils.definitions import ExitMessage
    from .utils.definitions import LogMessage


JSONTypes = bool | int | str | dict[Any, Any] | list[Any]


@dataclass
class Collector:
    """Data class for a collector."""

    name: str

    def start(self, color: bool) -> None:
        """Output start information to the console.

        Args:
            color: Whether to color the message
        """
        message = f"Collecting {self.name} information"
        information = f"{message:.<60}"
        ansi.working(color=color, message=information)

    def fail(self, color: bool, duration: float) -> None:
        """Output fail information to the console.

        Args:
            color: Whether to color the message
            duration: The duration of the collection
        """
        message = f"{self.name.capitalize()} information collection failed"
        information = f"{message:.<60}{duration:.2f}s"
        ansi.failed(color=color, message=information)

    def finish(self, color: bool, duration: float) -> None:
        """Output finish information to the console.

        Args:
            color: Whether to color the message
            duration: The duration of the collection
        """
        message = f"{self.name.capitalize()} information collected"
        information = f"{message:.<60}{duration:.2f}s"
        ansi.success(color=color, message=information)


@dataclass
class Diagnostics:
    """Data class for the diagnostics."""

    # pylint: disable=too-many-instance-attributes

    __WARNING__: dict[str, JSONTypes]
    basics: dict[str, JSONTypes]
    container_engines: dict[str, JSONTypes]
    execution_environment: dict[str, JSONTypes]
    initialization: dict[str, JSONTypes]
    local_system: dict[str, JSONTypes]
    logs: dict[str, JSONTypes]
    python_packages: dict[str, JSONTypes]
    settings: dict[str, JSONTypes]
    settings_file: dict[str, JSONTypes]


def register(collector: Collector) -> Callable[..., Any]:
    """Register a collector.

    Args:
        collector: The collector to register

    Returns:
        The decorator
    """

    def decorator(func: Callable[..., Any]) -> Callable[..., Any]:
        """Add the dunder collector to the func.

        Args:
            func: The function to decorate

        Returns:
            The decorated function
        """
        func.__collector__ = collector  # type: ignore[attr-defined]
        return func

    return decorator


DIAGNOSTIC_FAILURES = 0


class FailedCollectionError(Exception):
    """Exception for a failed collection."""

    def __init__(self, errors: dict[str, JSONTypes]) -> None:
        """Initialize the exception.

        Args:
            errors: The errors
        """
        super().__init__()
        self.errors = errors


def diagnostic_runner(func: Callable[..., Any]) -> Callable[..., Any]:
    """Wrap and run a collector.

    Args:
        func: The function to wrap

    Returns:
        The decorator
    """

    def wrapper(*args: Any, **kwargs: dict[str, Any]) -> Callable[..., Any]:
        """Wrap and run the collector.

        Args:
            *args: The positional arguments
            **kwargs: The keyword arguments

        Returns:
            The result of the function with elapsed or error information
        """
        global DIAGNOSTIC_FAILURES
        start = datetime.now(timezone.utc)
        color = args[0].color
        collector = func.__collector__  # type: ignore[attr-defined]
        collector.start(color=color)
        try:
            result = func(*args, **kwargs)
            duration = (datetime.now(timezone.utc) - start).total_seconds()
            collector.finish(color=color, duration=duration)
        except FailedCollectionError as error:
            # A collector exception, has data
            result = error.errors
            duration = (datetime.now(timezone.utc) - start).total_seconds()
            collector.fail(color=color, duration=duration)
            DIAGNOSTIC_FAILURES += 1
        except Exception as error:  # noqa: BLE001
            # Any other exception, has no data
            result = {"error": str(error) + "\n" + traceback.format_exc()}
            duration = (datetime.now(timezone.utc) - start).total_seconds()
            collector.fail(color=color, duration=duration)
            DIAGNOSTIC_FAILURES += 1
        result["duration"] = round(duration)
        return result

    return wrapper


class DiagnosticsCollector:
    """The diagnostics collector."""

    WARNING = "The following output may contain sensitive data, please review it carefully."

    def __init__(
        self,
        args: ApplicationConfiguration,
        messages: list[LogMessage],
        exit_messages: list[ExitMessage],
    ) -> None:
        """Initialize the ShowTech class.

        Args:
            args: The current settings
            messages: The messages to log
            exit_messages: The exit messages to log
        """
        self._args = args
        self.color = args.display_color and stdout.isatty()
        self._messages = messages
        self._exit_messages = exit_messages

    @property
    def registered(self) -> Iterator[Callable[..., Any]]:
        """Return the registered diagnostics.

        Returns:
            The registered diagnostics
        """
        return (getattr(self, f) for f in dir(self) if hasattr(getattr(self, f), "registration"))

    def run(self) -> None:
        """Collect as much information as possible about everything and dump to a json file."""
        ansi.warning(color=self.color, message=self.WARNING)
        diagnostics = Diagnostics(
            __WARNING__=self._warning(),
            basics=self._basics(),
            container_engines=self._container_engines(),
            execution_environment=self._execution_environment(),
            initialization=self._initialization(),
            local_system=self._local_system(),
            logs=self._log_collector(),
            python_packages=self._python_packages(),
            settings=self._settings(),
            settings_file=self._settings_file(),
        )

        time = now_iso("local")
        file_name = f"diagnostics-{time}.json"
        path = f"{Path.home()}/{file_name}"
        mode = 0o600
        write_diagnostics_json(path, mode, asdict(diagnostics))
        message = f"\nDiagnostics written to: {path}"

        if DIAGNOSTIC_FAILURES > 0:
            ansi.warning(color=self.color, message=message)
        else:
            ansi.success(color=self.color, message=message)
        sys.exit(0)

    @diagnostic_runner
    @register(Collector(name="warning"))
    def _warning(self) -> dict[str, JSONTypes]:
        """Add a warning.

        Returns:
            The warning
        """
        return {"message": self.WARNING}

    @diagnostic_runner
    @register(Collector(name="basic"))
    def _basics(self) -> dict[str, JSONTypes]:
        """Add basic information.

        Returns:
            The basic information
        """
        return {
            "application_name": self._args.application_name,
            "application_version": str(self._args.application_version),
            "action_packages": list(self._args.internals.action_packages),
            "cache_path": str(self._args.internals.cache_path),
            "collection_doc_cache": str(self._args.internals.collection_doc_cache),
            "original_command": shlex_join(sys.argv),
            "settings_file_path": str(self._args.internals.settings_file_path),
            "settings_source": str(self._args.internals.settings_source),
        }

    @diagnostic_runner
    @register(Collector(name="container engines"))
    def _container_engines(self) -> dict[str, JSONTypes]:
        """Add container engines.

        Returns:
            The container engines
        """
        commands = [
            Command(identity="podman", command="podman --version", post_process=lambda c: c),
            Command(identity="docker", command="docker --version", post_process=lambda c: c),
        ]
        CommandRunner().run_single_process(commands)
        engines: dict[str, JSONTypes] = {}
        for command in commands:
            engines[command.identity] = {
                "return_code": command.return_code,
                "selected": bool(self._args.container_engine == command.identity),
                "stdout": command.stdout,
                "stderr": command.stderr,
            }
        return engines

    @diagnostic_runner
    @register(Collector(name="execution environment"))
    def _execution_environment(self) -> dict[str, JSONTypes]:
        """Add execution environment information.

        Raises:
            FailedCollectionError: If the collection process fails

        Returns:
            The execution environment information
        """
        if self._args.entry("container_engine").value.source is Constants.DEFAULT_CFG:
            return {"errors": "No container engine available or found"}

        details, errors, return_code = introspector.run(
            image_name=self._args.execution_environment_image,
            container_engine=self._args.container_engine,
        )
        details = {"details": details, "errors": errors, "return_code": return_code}
        if errors or not details:
            raise FailedCollectionError(details)
        return details

    @diagnostic_runner
    @register(Collector(name="initialization"))
    def _initialization(self) -> dict[str, JSONTypes]:
        """Add initialization information.

        Returns:
            The initialization information
        """
        return {
            "messages": [msg.message for msg in self._messages],
            "exit_messages": [msg.message for msg in self._exit_messages],
        }

    @diagnostic_runner
    @register(Collector(name="log"))
    def _log_collector(self) -> dict[str, JSONTypes]:
        """Add log collector information.

        Returns:
            The log collector information
        """
        logs: list[JSONTypes] = []
        cwd_log = Path("./ansible-navigator.log")
        if cwd_log.exists():
            contents = cwd_log.read_text(encoding="utf-8").splitlines()
            log = {
                "name": str(cwd_log),
                "contents": contents,
                "date": cwd_log.stat().st_mtime,
            }
            logs.append(log)
        settings_log = Path(self._args.log_file)
        if cwd_log != settings_log and settings_log.exists():
            contents = settings_log.read_text(encoding="utf-8").splitlines()
            log = {
                "name": str(settings_log),
                "contents": contents,
                "date": settings_log.stat().st_mtime,
            }
            logs.append(log)
        return {"found": bool(logs), "logs": logs}

    @diagnostic_runner
    @register(Collector(name="local system"))
    def _local_system(self) -> dict[str, JSONTypes]:
        """Add local system information.

        Raises:
            FailedCollectionError: If the collection process fails

        Returns:
            The local system information
        """
        results = image_introspect.main(serialize=False)
        if not results:
            raise FailedCollectionError({})
        errors = results.get("errors")
        if errors:
            if not isinstance(errors, dict):
                msg = "Unexpected data"
                raise RuntimeError(msg)
            raise FailedCollectionError(errors)  # type: ignore
        return {"details": results}

    @diagnostic_runner
    @register(Collector(name="python packages"))
    def _python_packages(self) -> dict[str, JSONTypes]:
        """Add python packages information.

        Returns:
            The python packages information
        """
        pkgs = importlib_metadata.packages_distributions()
        meta: dict[str, Any] = {}
        for pkg_names in pkgs.values():
            for pkg_name in pkg_names:
                if pkg_name not in meta:
                    meta[pkg_name] = importlib_metadata.metadata(pkg_name).json
                    spec = find_spec(pkg_name)
                    if spec:
                        meta[pkg_name]["location"] = spec.origin
                    meta[pkg_name]["requires"] = importlib_metadata.distribution(pkg_name).requires
        return meta

    @diagnostic_runner
    @register(Collector(name="settings"))
    def _settings(self) -> dict[str, JSONTypes]:
        """Add settings information.

        Returns:
            The settings information
        """
        return {
            "effective": to_effective(self._args),
            "sources": to_sources(self._args),
        }

    @diagnostic_runner
    @register(Collector(name="settings file"))
    def _settings_file(self) -> dict[str, JSONTypes]:
        """Add settings file information.

        Returns:
            The settings file information
        """
        contents: dict[str, JSONTypes] = {}
        if self._args.internals.settings_file_path:
            text = Path(self._args.internals.settings_file_path).read_text(encoding="utf-8")
            contents = yaml.load(text, Loader=Loader)
        return {"contents": contents}



from __future__ import annotations

import json
import os

from pathlib import Path
from typing import Any
from typing import NamedTuple
from typing import TypeVar

from .compiler import Compiler
from .fchainmap import FChainMap
from .reg import _Reg
from .reg import make_reg
from .rules import Rule
from .rules import _Rule
from .utils import uniquely_constructed


T = TypeVar("T")

SCOPE_UNKNOWN = "source.unknown"


@uniquely_constructed
class Grammar(NamedTuple):
    scope_name: str
    repository: FChainMap[str, _Rule]
    patterns: tuple[_Rule, ...]

    @classmethod
    def make(cls, data: dict[str, Any]) -> Grammar:
        scope_name = data["scopeName"]
        if "repository" in data:
            # this looks odd, but it's so we can have a self-referential
            # immutable-after-construction chain map
            repository_dct: dict[str, _Rule] = {}
            repository = FChainMap(repository_dct)
            for k, dct in data["repository"].items():
                repository_dct[k] = Rule.make(dct, repository)
        else:
            repository = FChainMap()
        patterns = tuple(Rule.make(d, repository) for d in data["patterns"])
        return cls(scope_name=scope_name, repository=repository, patterns=patterns)


class Grammars:
    def __init__(self, *directories: str) -> None:
        """Initialize an instance of Grammars.

        Args:
            *directories: A tuple of strings, each a directory in which
                grammar files can be found
        """
        self._scope_to_files = {
            Path(filename).stem: Path(directory) / filename
            for directory in directories
            if Path(directory).exists()
            for filename in sorted(os.listdir(directory))  # noqa: PTH208
            if filename.endswith(".json")
        }

        unknown_grammar = {"scopeName": SCOPE_UNKNOWN, "patterns": []}
        self._raw = {SCOPE_UNKNOWN: unknown_grammar}
        self._file_types: list[tuple[frozenset[str], str]] = []
        self._first_line: list[tuple[_Reg, str]] = []
        self._parsed: dict[str, Grammar] = {}
        self._compiled: dict[str, Compiler] = {}

    def _raw_for_scope(self, scope: str) -> dict[str, Any]:
        try:
            return self._raw[scope]
        except KeyError:
            pass

        grammar_path = Path(self._scope_to_files.pop(scope))
        with grammar_path.open(encoding="UTF-8") as f:
            ret = self._raw[scope] = json.load(f)

        file_types = frozenset(ret.get("fileTypes", ()))
        first_line = make_reg(ret.get("firstLineMatch", "$impossible^"))

        self._file_types.append((file_types, scope))
        self._first_line.append((first_line, scope))

        return ret

    def grammar_for_scope(self, scope: str) -> Grammar:
        try:
            return self._parsed[scope]
        except KeyError:
            pass

        raw = self._raw_for_scope(scope)
        ret = self._parsed[scope] = Grammar.make(raw)
        return ret

    def compiler_for_scope(self, scope: str) -> Compiler:
        try:
            return self._compiled[scope]
        except KeyError:
            pass

        grammar = self.grammar_for_scope(scope)
        ret = self._compiled[scope] = Compiler(grammar, self)
        return ret

    def blank_compiler(self) -> Compiler:
        return self.compiler_for_scope(SCOPE_UNKNOWN)

    def compiler_for_file(self, filename: str, first_line: str) -> Compiler:
        # didn't find it in the fast path, need to read all the json
        for k in tuple(self._scope_to_files):
            self._raw_for_scope(k)

        _, _, ext = Path(filename).name.rpartition(".")
        for extensions, scope in self._file_types:
            if ext in extensions:
                return self.compiler_for_scope(scope)

        for reg, scope in self._first_line:
            if reg.match(first_line, 0, first_line=True, boundary=True):
                return self.compiler_for_scope(scope)

        return self.compiler_for_scope(SCOPE_UNKNOWN)
